/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include "str-lib.h"
#include "rafutil.h"
#include "dafstype.h"
#include "entity.h"
#include "dafs.h"
#include "smemory.h"

/* Macro used throughout this module to determine byte ordering */
static int2 __s=1;
#define INTEL_BYTE_ORDER ( *(char*) &__s == 1 )

static 
iDAFSError DeleteProperty(iPropPtr *p, const char *name)
{
	iPropPtr pp;
	iPDataPtr pd;
	
	if (!p || !name) return NotEnoughData;
	if (pp = svbmapSeek(*p,name)) {
		if (pd = pp->data) {
			switch(pd->type) {
			  case iPString:
				if (pd->val.s)
					smfree(pd->val.s);
				break;
			  case iPUString:
				if (pd->val.us)
					smfree(pd->val.us);
				break;
			  case iPUserData:
			  case iPUserIData:
				if (pd->size && pd->val.v)
					smfree(pd->val.v);
			  case iPUserPtr:
				if(pd->callback)
					pd->callback(pd->val.v);
				break;
			default:
				break;
			}
			smfree(pd);
		}
		svbmapDeleteRecord(p,pp);
	}

	return DAFSOK;
}

static iDAFSError 
_SetProp(iPropPtr *p, const char *name, iPTag type, 
		 uint2 size, iUserPtrFree callback, void *data)
{
	int flag = 0;
	iPDataPtr pd;
	iPropPtr pp;
	iDAFSError e;
	char *hold=0;
	void *temp;

	if (!p) return NotEnoughData;

	/* pre-allocate space for data, so if memory runs out
	   clean up is trival */
	switch(type){
	  case iPString:
		if(!data) break;
		temp = smalloc(strlen(data)+1);
		if(!temp) return MallocFailError;
		break;

	  case iPUString:
		if(!data) break;
		temp = smalloc(sizeof(iUCode)*(ucslen(data)+1));
		if(!temp) return MallocFailError;
		break;
		
	  case iPUserIData:
	  case iPUserData:
		if(!size) break;
		temp = smalloc(size);
		if(!temp) return MallocFailError;
		break;
	}

	if (!(pp=svbmapSeek(*p,name))) {
		pd = (iPDataPtr)smalloc(sizeof(iPData));
		if (!pd) return MallocFailError;
		pd->val.s = 0;
		pd->type = 0;
		pd->size = 0;
		pd->callback = 0;
		if ((e = svbmapSet(p,name,pd)) != DAFSOK){
			e=MallocFailError;
			smfree(pd);
			return e;
		}
		flag = 1;
	} else {
		pd = pp->data;
		/* bogus should always have data size */
		if ((pd->type == iPUserData)||(pd->type == iPUserIData) /*  && pd->size */)
			if (pd->val.v) {
				hold = (char*)pd->val.v;
				pd->val.v = 0;
			}
		if (pd->type == iPString && pd->val.s) {
			hold = pd->val.s;
			pd->val.s = 0;
		}
		if (pd->type == iPUString && pd->val.us) {
			hold = (char*)pd->val.us;
			pd->val.us = 0;
		}
		pd->size = size;
	}
	
	switch(type) {
	  case iPString:
		pd->val.s = (char *) temp;
		strcpy(pd->val.s, data);
		pd->size = strlen(pd->val.s) + 1;
		break;

	  case iPUString:
		pd->val.us = (iUCode *) temp;
		ucscpy(pd->val.us, data);
		pd->size = (ucslen(pd->val.us) + 1)*sizeof(iUCode);
		break;
		
	  case iPUserData:
		pd->val.v = (char *) temp;
		pd->size = size;
		if (data) memcpy(pd->val.v, data, size);
		break;

	  case iPUserIData:
		pd->val.v = (char *) temp;
		if (data) memcpy(pd->val.v, data, size);
		pd->size = size;
		break;

	  case iPUserPtr:
		pd->val.v = data;
		pd->callback = callback;
		break;

	  case iPLong:
	  case iPBool:
		pd->val.i = (int4)data;
		pd->size=sizeof(int4); /* force it in case zero is passed to func. */
		break;

	  default: 
		break; 
	}
	pd->type = type;
	if (hold) smfree(hold);
	
	return DAFSOK;
}


iDAFSError SetTypeProperty(iTypeIndex itype,const char *name,     
			   iPTag type,uint2 size,void *data,iUserPtrFree callback,int4 fromdisk)
{
	if (itype >= i_GlobalType.size)
		return 0;
	if (INTEL_BYTE_ORDER && !fromdisk && type==iPUserData)
		type=iPUserIData;
	return _SetProp(&(i_GlobalType.table[itype].props),name,type,size,callback,data);
}

Dll iDAFSError APIENTRY
i_SetSessionBoolProp(iTypeIndex itype,const char *name,   
				  int data)
{
	return SetTypeProperty(itype,name,iPBool,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetSessionLongProp(iTypeIndex itype,const char *name,   
				  int4 data)
{
	return SetTypeProperty(itype,name,iPLong,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetSessionStringProp(iTypeIndex itype,const char *name,   
				  char *data)
{
	return SetTypeProperty(itype,name,iPString,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetSessionUStringProp(iTypeIndex itype,const char *name,   
				  iUCode *data)
{
	return SetTypeProperty(itype,name,iPUString,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetSessionUserDataProp(iTypeIndex itype,const char *name,   
				  void *data, uint2 size)
{
	return SetTypeProperty(itype,name,iPUserData,size,data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetSessionUserPtrProp(iTypeIndex itype,const char *name,   
						void *data, iUserPtrFree callback)
{
	return SetTypeProperty(itype,name,iPUserPtr,0,data,callback,0);
}

iDAFSError
SetProperty(iEntityPtr entity, const char *name, iPTag type, 
			uint2 size, void *data, iUserPtrFree callback, int fromdisk)
{
	iDAFSError err;
		
	if (!entity || !entity->data) return NotEnoughData;
	if (gCallBack)
		MatrCallBack(mWillUpdate,gUserData,entity);
	if (INTEL_BYTE_ORDER && !fromdisk && type==iPUserData)
		type=iPUserIData;
	if( err = _SetProp(&(entity->data->props),name,type,size,callback,data)) 
		return err;
	if (gCallBack)
		MatrCallBack(mUpdate,gUserData,entity);
	return err;
}

static iDAFSError 
i_SetProperty(iEntityPtr entity, const char *name,  
			  iPTag type,int2 size,void *data)
{
	if (!entity || !name) return NotEnoughData;
	return SetProperty(entity,name,type,size,data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetBoolProp(iEntityPtr entity, const char *name,  
			  int data)
{
	if (!entity || !name) return NotEnoughData;
	return SetProperty(entity,name,iPBool,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetLongProp(iEntityPtr entity, const char *name,  
			  int4 data)
{
	if (!entity || !name) return NotEnoughData;
	return SetProperty(entity,name,iPLong,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetStringProp(iEntityPtr entity, const char *name,  
			  char *data)
{
	if (!entity || !name) return NotEnoughData;
	return SetProperty(entity,name,iPString,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetUStringProp(iEntityPtr entity, const char *name,  
			  iUCode *data)
{
	if (!entity || !name) return NotEnoughData;
	return SetProperty(entity,name,iPUString,0,(void*)data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetUserDataProp(iEntityPtr entity, const char *name,  
			  void *data, uint2 size)
{
	if (!entity || !name) return NotEnoughData;
	return SetProperty(entity,name,iPUserData,size,data,NULL,0);
}

Dll iDAFSError APIENTRY
i_SetUserPtrProp(iEntityPtr entity, const char *name,  
				 void *data, iUserPtrFree callback)
{
	if (!entity || !name) return NotEnoughData;
	return SetProperty(entity,name,iPUserPtr,0,data,callback,0);
}

Dll iDAFSError APIENTRY
i_MoveProp(iEntityPtr entity,iEntityPtr toen,const char *name)  
{
	iDAFSError e;

	if (!name || !entity || !toen || !entity->data->props)
		return NotEnoughData;
	
	if (!svbmapContainsP(entity->data->props,name))
		return NonExistentProp;
	
	if (gCallBack) {
		MatrCallBack(mWillUpdate,gUserData,entity);
		MatrCallBack(mWillUpdate,gUserData,toen);
	}
	if (svbmapContainsP(toen->data->props, name))
		DeleteProperty(&toen->data->props,name);
	e = svbmapSet(&toen->data->props,name,svbmapGet(entity->data->props,name)->data);
	if (e != DAFSOK){
		e=MallocFailError;
		return e;
	}
	svbmapDelete(&entity->data->props,name);
	if (gCallBack) {
		MatrCallBack(mUpdate,gUserData,entity);
		MatrCallBack(mUpdate,gUserData,toen);
	}
	return DAFSOK;
}

static iDAFSError 
i_MoveAllPropsPtr(iPropPtr *from_p, iPropPtr *to_p)  
{	
	iPropPtr mp;
	iDAFSError e;

	if (!from_p || !to_p) return NotEnoughData;
	if (!*from_p) return DAFSOK;

	for (mp=svbmapFirst(*from_p); mp; mp=svbmapFirst(*from_p))  {
		/* if destination property table already has a property
		   with this name, delete the property and table entry */
		if (svbmapContainsP(*to_p,mp->name))
			DeleteProperty(to_p,mp->name);
		if ((e = svbmapSet(to_p,mp->name,mp->data)) != DAFSOK){
			e=MallocFailError; 
			return e;
		}
		svbmapDelete(from_p,mp->name);
	}
	svbmapFree(from_p);
	return DAFSOK;
}	

Dll iDAFSError APIENTRY
i_MoveAllProps(iEntityPtr from_en, iEntityPtr to_en)  
{	
	iDAFSError e;

	if (!from_en || !to_en) return NotEnoughData;
	if (!from_en->data->props) return DAFSOK;
	if ((e=i_MoveAllPropsPtr(&from_en->data->props, &to_en->data->props)))
		return e;
	return DAFSOK;
}	

Dll iDAFSError APIENTRY
i_CopyAllProps(iEntityPtr from_en, iEntityPtr to_en)  
{	
	svbmap *mp;
	iDAFSError e;

	if (!from_en || !to_en) return NotEnoughData;
	if (!from_en->data->props) return DAFSOK;

	for (mp=svbmapFirst(from_en->data->props); mp; mp=svbmapNext(from_en->data->props,mp))  {
		if (e=i_SetProperty(to_en,mp->name,((iPDataPtr)mp->data)->type,
			     ((iPDataPtr)mp->data)->size,
			     ((iPDataPtr)mp->data)->val.v))
			return e;
	}
	return DAFSOK;
}	


static iPropPtr 
i_FindPropPtr(const iPropPtr p, const char *name)
{
	iPropPtr mp;

	if (!p || !name || !(mp = svbmapSeek(p,name))) return 0;
	return mp;
}
static void
getVoidData(iPropPtr p, void **data){
	void *void_data;
	
	iPDataPtr pd = p->data;
	switch(pd->type) {
	  case iPString:
		void_data = pd->val.s;
		break;
	  case iPUString:
		void_data = pd->val.us;
		break;
	  case iPUserData:
		void_data = pd->val.v;
		break;
	  case iPUserIData:
		void_data = pd->val.v;
		break;
	  case iPUserPtr:
		void_data = pd->val.v;
		break;
	  case iPLong:
	  case iPBool:
		void_data = pd->val.i;
		break;
	  default: 
		break; 
	}
	*data = void_data;
}	

static void
i_DisposePropPtr(const iPropPtr *p, const char *name)
{
	DeleteProperty((iPropPtr*)p, name);
}


Dll iDAFSError APIENTRY
i_ChangePropName(iEntityPtr entity,const char *from_name, const char *to_name)
{
	iPropPtr p;
	void *void_data;
	iPDataPtr data ;
	iDAFSError error = DAFSOK;
	iUserPtrFree *callback = 0;
	
	if (!from_name || !to_name) 
		return NotEnoughData;
	if (entity && (p = i_FindPropPtr(entity->data->props,from_name))){
		getVoidData(p, &void_data);
		data = p->data;
		error = SetProperty(entity, to_name, data->type, data->size,
					void_data, data->callback,0);
		i_DisposePropPtr(&entity->data->props, from_name);
	} else if (p = i_FindPropPtr(i_GlobalType.table[entity->data->type].props,from_name)){
		getVoidData(p, &void_data);
		data = p->data;
		error = SetTypeProperty(i_GetEntityType(entity), to_name, data->type,
						data->size, void_data, data->callback,0);
		i_DisposePropPtr(i_GlobalType.table[entity->data->type].props, from_name);
	}
	return error;
}
	
Dll iPropPtr APIENTRY
i_FindProp(const iEntityPtr entity, const char *name)  
{
	iPropPtr p;

	if (!name || !entity || !entity->data) return 0;
	if (!(p = i_FindPropPtr(entity->data->props,name)))
		p = i_FindPropPtr(i_GlobalType.table[entity->data->type].props,name);
	return p;
}

Dll iPropPtr APIENTRY
i_FindPropNoTypes(const iEntityPtr entity, const char *name)  
{
	return i_FindProp(entity,name);
}

Dll void APIENTRY
i_DisposeProp(const iEntityPtr entity, const char *name)  
{
	if (!name || !entity || !entity->data->props ||
            !svbmapContainsP(entity->data->props, name)) return;
	if (gCallBack) MatrCallBack(mWillUpdate,gUserData,entity);
	i_DisposePropPtr(&entity->data->props, name);
	if (svbmapEmptyP(entity->data->props)) 
		svbmapFree(&entity->data->props);
	if (gCallBack) MatrCallBack(mUpdate,gUserData,entity);
}

void
i_DisposeAllPropsPtr(const iPropPtr  *p)
{
	iPropPtr pp;

	if (!p) return;
	for (pp=svbmapFirst(*p); pp; pp=svbmapFirst(*p)) {
		iPDataPtr pd = pp->data;
		switch(pd->type) {
        	case iPString:
                	if (pd->val.s)
                        	smfree(pd->val.s);
                	break;
        	case iPUString:
                	if (pd->val.us)
                        	smfree(pd->val.us);
                	break;
        	case iPUserData:
        	case iPUserIData:
                	if (pd->size && pd->val.v)
                        	smfree(pd->val.v);
                    break;
            case iPUserPtr:
                    if(pd->callback)
                            pd->callback(pd->val.v);
                    break;
        	default:
                	break;
		}
		smfree(pd);
		svbmapDeleteRecord((iPropPtr*)p,pp);
	}
}

Dll void APIENTRY
i_DisposeAllProps(const iEntityPtr entity)
{
	if (!entity || !entity->data || !entity->data->props) return;
	if (gCallBack) MatrCallBack(mWillUpdate,gUserData,entity);
	i_DisposeAllPropsPtr(&entity->data->props);
	if (gCallBack) MatrCallBack(mUpdate,gUserData,entity);
}


Dll void APIENTRY
i_DisposeSessionProp(iTypeIndex index,const char *name)     
{
	if (!name || !svbmapContainsP(i_GlobalType.table[index].props, name)) return;
	i_DisposePropPtr(&i_GlobalType.table[index].props, name);
}

Dll void APIENTRY
i_DisposeAllSessionProps(void)
{
	int4 i;

	for (i=0; i< i_GlobalType.size; i++)
		i_DisposeAllPropsPtr(&i_GlobalType.table[i].props);
}

Dll iDAFSError APIENTRY
i_GetUserDataProp(const iEntityPtr entity, const char *name, void **d,
				  int4 *size, iUserDataProc ConvertProc)
{
	iPropPtr p;
	iPDataPtr pd;

	if(d) *d = 0;
	if (!name || !entity) return DAFSOK;
	if (!(p = i_FindProp(entity,name))) return NonExistentProp;
	pd = p->data;
	if ((pd->type != iPUserData)&&(pd->type != iPUserIData)) 
		return WrongPropType;
	/*  CAN BE REMOVED AFTER FILES ARE CONVERTED  */
	if (d&&(!strcmp("PBB",name)|| !strcmp("LPBB",name))&&(pd->size>sizeof(iBox))){
		/* Exit and Die */
		halt("*** bad PBB size, run fixpbb to convert RafFile **\n");
	}
	else {
		if (ConvertProc){
			if (INTEL_BYTE_ORDER) {	/* bytes should be in Intel order  */
				if (pd->type == iPUserData){
					ConvertProc(pd->val.v);
					pd->type = iPUserIData;
				}
			}
			else {			/* bytes should be in straight order */
				if (pd->type == iPUserIData) {  
					ConvertProc(pd->val.v);
					pd->type = iPUserData;
				}
			}
		}
	}		
	if (d) *d = pd->val.v;
	if (size) *size = pd->size;

	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_GetUserPtrProp(const iEntityPtr entity,const char *name, void **d,
		iUserPtrFree *callp)  
{
	iPropPtr p;
	iPDataPtr pd;

	if (!name || !entity) return DAFSOK;
	if (d) *d = NULL;
	if (!(p = i_FindProp(entity,name)))
		return NonExistentProp;
	pd = p->data;
	if (pd->type != iPUserPtr)
		return WrongPropType;
	if (d) *d =  pd->val.v;
	if (callp) *callp = pd->callback;
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_GetLongProp(const iEntityPtr entity,const char *name,int4 *ip)  
{
	iPropPtr p;
	iPDataPtr pd;

	if (!name || !entity) return DAFSOK;
	if (!(p = i_FindProp(entity,name))) return NonExistentProp;
	pd = p->data;
	if (pd->type != iPLong) return WrongPropType;
	if (ip) *ip = pd->val.i;
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_GetStringProp(const iEntityPtr entity,const char *name,char **ip)  
{
	iPropPtr p;
	iPDataPtr pd;

	if (!name || !entity) return DAFSOK;
	if (ip) *ip = NULL;
	if (!(p = i_FindProp(entity,name))) return NonExistentProp;
	pd = p->data;
	if (pd->type != iPString) return WrongPropType;
	if (ip) *ip = pd->val.s;
	return DAFSOK;
}

Dll iDAFSError APIENTRY
i_GetUStringProp(const iEntityPtr entity,const char *name,iUCode **ip)  
{
	iPropPtr p;
	iPDataPtr pd;

	if (!name || !entity) return DAFSOK;
	if (ip) *ip = NULL;
	if (!(p = i_FindProp(entity,name))) return NonExistentProp;
	pd = p->data;
	if (pd->type != iPUString) return WrongPropType;
	if (ip) *ip = pd->val.us;
	return DAFSOK;
}


Dll iDAFSError APIENTRY
i_GetBoolProp(const iEntityPtr entity,const char *name,int4 *ip)  
{
	iPropPtr p;
	iPDataPtr pd;

	if (!name || !entity) return DAFSOK;
	if (ip) *ip = 0;
	if (!(p = i_FindProp(entity,name))) return NonExistentProp;
	pd = p->data;
	if (pd->type != iPBool) return WrongPropType;
	if (ip) *ip = pd->val.i;
	return DAFSOK;
}

/* not static because referenced in file.c */

iPropPtr
i_FirstPropPtr(const iPropPtr pm)
{
	if (!pm) return 0;
	return (iPropPtr) svbmapFirst(pm);
}

iPropPtr
i_NextPropPtr(const iPropPtr pm, const iPropPtr prev)
{
	if (!pm) return 0;
	return svbmapNext(pm, prev);
}

Dll iPropPtr APIENTRY
i_FirstProp(const iEntityPtr entity)
{
	if (!entity || !entity->data) return 0;
	return i_FirstPropPtr(entity->data->props);
}

Dll iPropPtr APIENTRY
i_NextProp(const iEntityPtr entity, const iPropPtr prev)
{
	if (!entity || !entity->data || !prev) return 0;
	return i_NextPropPtr(entity->data->props, prev);
}

Dll int4 APIENTRY 
i_GetPropSize(const iPropPtr p)  
{
	iPData *data;
	if (!p) return 0;

	data = (iPData *) p->data;
	return data->size;
}

Dll const char * APIENTRY
i_GetPropName(const iPropPtr p)  
{
	if (!p) return 0;
	return p->name;
}

Dll iPTag APIENTRY
i_GetPropType(const iPropPtr p)  
{
	iPData *data;

	if (!p) return 0;
	data = (iPData *) p->data;

	return data->type;
}


Dll const char * APIENTRY
i_GetComment(const iEntityPtr entity)  
{
	iPropPtr p;
	iPData *data;

	if (!(p = i_FindProp(entity,"comment"))) return 0;
	data = (iPData *) p->data;

	return data->val.s;
}

Dll iDAFSError APIENTRY
i_SetComment(iEntityPtr entity,const char *comment)  
{
	iDAFSError err;
	
	if (!comment || !*comment) {
		i_DisposeProp(entity, "comment");
		return NotEnoughData;
	}
	err=i_SetStringProp(entity, "comment",(char*)comment);
	
	return err;
}




#if 0
iDAFSError 
i_MakeProp(iPropPtr *prop, const char *name, iPTag type, uint2 size, void *data)
{
	/* check for duplicate property */
	if (*prop && i_FindPropPtr(*prop, name))
		return DAFSOK;
	else
		return _SetProp(prop, name, type, size, NULL, data);
}
#endif

#if 0
iDAFSError 
i_AttachProps(iEntityPtr entity, iPropPtr *prop)
{
	iDAFSError e;

	if (!entity || !entity->data ) return NotEnoughData;
	if (!prop) return DAFSOK;

	if((e=i_MoveAllPropsPtr(prop,&entity->data->props))) return e;
		
	return DAFSOK;
}
#endif
