
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "c-lib.h"
#include "str-lib.h"

#include "rafutil.h"
#include "entity.h"
#include "rlimage.h"

#include "dafs.h"
#include "smemory.h"
#include "tiffio.h"

#define CHUNK 8192
#define MAXWID 32767

/* static data size is 256 bytes */

static unsigned char swabtab[256] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
};


static void swabit(unsigned char *data,uint2 len)
{
	for (;len--;data++)
		*data = swabtab[*data];
}



static int4 
tifffillline(RafFile  in, char *inbuf, int2 need, int4 *count)
{
	int4 e;
	e = TIFFReadScanline((TIFF *) in, (unsigned char *) inbuf, (*count)++, 0);
	if (e < 0)
		return e;
	return 0;
}


static iDAFSError 
rawtiff(iImagePtr *npp,TIFF * in, iDAFSCompType comp,int2 w,
		int2 h, int4 swab,int invert)
{
	int4 cur=0,rowsper;
	uint2 i,nchunks;
	long *strips,buflen,len,got;
 	iImagePtr np=0;
	iDAFSError e = DAFSOK;
	
	np = newrawimage(w, h,comp);
	if (!np) { e = MallocFailError; goto error; }
	
	rowsper = 0;
	TIFFGetField(in, TIFFTAG_ROWSPERSTRIP, &rowsper);
	nchunks = TIFFNumberOfStrips(in);
	TIFFGetField(in, TIFFTAG_STRIPBYTECOUNTS, &strips);
	
	for (i=0;i<nchunks;i++) {
		buflen = strips[i];
		got = 0;
		while (buflen) {
			len = buflen;
			if (len > CHUNK) len = CHUNK;

			if (e = newchunk(np,(uint2)len))
				goto error;
			if (TIFFReadRawStripSeek(in,i,np->d.rw.chunks[cur].data, (unsigned)len,got)!=(int)len)
				{e = ReadError; goto error; }
			if (swab)
				swabit((unsigned char *)np->d.rw.chunks[cur].data,(uint2)len);
			cur++;
			buflen -= len;
			got += len;
		}
	}
	np->d.rw.invert = invert;
	if ((nchunks > 1) && (rowsper > 1) && (rowsper < np->height)) {
		if (e = i_uncompressimage(np,rowsper))
			goto error;
	}
	*npp = np;
	np = 0;
error:
	if (np) {
		if (np->type == Raw) freeimage(np);
		else if (np->type == Image) freeimage(np);
	}
	return e;
}


iDAFSError 
tiffread(iImagePtr * npp, const char *fn)
{
	TIFF *in;
	int2 h, w;
	int4 hl,wl,opt;
	short s, comp,swab,invert;
	int2 res=0;
	float f_res=(float)0.0;
	iDAFSError e = DAFSOK;
	iImagePtr np=0;
	
	in = TIFFOpen((char *)fn, "rb");
	if (!in) { e = ReadError; goto error; }
	TIFFGetField(in, TIFFTAG_SAMPLESPERPIXEL, &s);
	if (s != 1) { e = ReadError; goto error; }
	TIFFGetField(in, TIFFTAG_BITSPERSAMPLE, &s);
	if (s != 1) { e = ReadError; goto error; }

	TIFFGetField(in, TIFFTAG_IMAGELENGTH, &hl);
	if (hl>MAXWID) { e = BadBoundsError; goto error; }
	h=(int2)hl;
	
	TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &wl);
	if (wl > MAXWID) { e = BadBoundsError; goto error; }
	w=(int2)wl;

	TIFFGetField(in, TIFFTAG_COMPRESSION, &comp);
	TIFFGetField(in, TIFFTAG_FILLORDER,&swab);
	TIFFGetField(in, TIFFTAG_XRESOLUTION, &f_res);
	TIFFGetField(in, TIFFTAG_PHOTOMETRIC, &invert);
	invert = invert == PHOTOMETRIC_MINISBLACK;
	res = (int2) f_res;
	
	switch (comp) {
	case 3:
		TIFFGetField(in,TIFFTAG_GROUP3OPTIONS,&opt);
		if (opt & GROUP3OPT_2DENCODING) comp = CCITT32D + 2;
	case 4:
		if (e = rawtiff(&np,in, comp-2,w, h,swab == FILLORDER_LSB2MSB,invert))
			goto error;
		break;
		
	case 1: /* doesn't handle wrong fillorder */
		if (e = do0d(&np,(RafFile ) in, w, h, tifffillline,invert))
			goto error;
		break;
	case 5: /* lempel zif & walsh: we can't handle it */
	default:
		e = ReadError;
		goto error;
	}
	if (!res) res = 300;
	i_SetImageRes(np, res);
	*npp = np;
	np = 0;
error:
	if (in) TIFFClose(in);
	if (np) freeimage(np);
	return e;
}


static iDAFSError 
writeuntiff(iImagePtr sp, TIFF * fo)
{
	iDAFSError e=DAFSOK;
	char *buf=0;
	int4 w8, i;

	w8 = (sp->width + 7) >> 3;
	buf = smalloc(sp->width * 2);
	if (!buf) { e = MallocFailError; goto error; }
	for (i = 0; i < sp->height; i++) {
		Run2Ras((short *) sp->d.im.rl[i].run, buf);
		if (TIFFWriteScanline(fo, (unsigned char*)buf, i, 0) < 0)
			{ e = WriteError; goto error; }
	}
error:
	if (buf) smfree(buf);
	return e;
}


static iDAFSError writerawtiff(iImagePtr sp, TIFF *fo)
{
	int4 i;
	iDAFSError e = DAFSOK;

	TIFFSetField(fo, TIFFTAG_COMPRESSION, (unsigned short) sp->d.rw.comp + 2);
	for (i=0;i<sp->d.rw.nchunks;i++) {
		if (TIFFWriteRawStrip(fo, 0,(unsigned char*)sp->d.rw.chunks[i].data, 
						sp->d.rw.chunks[i].len) < 0)
			{ e = WriteError; goto error; }
	}
error:
	return e;
}

iDAFSError writetiff(iImagePtr sp, const char *fn, int4 comp)
{
	iDAFSError e = DAFSOK;
	TIFF *fo=0;

	fo = TIFFOpen((char *)fn, "w");
	if (!fo) { e = WriteError; goto error; }
		
	TIFFSetField(fo, TIFFTAG_IMAGELENGTH, sp->height);
	TIFFSetField(fo, TIFFTAG_IMAGEWIDTH, sp->width);
	TIFFSetField(fo, TIFFTAG_PLANARCONFIG, 1);
	TIFFSetField(fo, TIFFTAG_SAMPLESPERPIXEL, 1);
	TIFFSetField(fo, TIFFTAG_BITSPERSAMPLE, 1);
	TIFFSetField(fo, TIFFTAG_FILLORDER,FILLORDER_MSB2LSB);
	if (sp->dpi) {
		TIFFSetField(fo, TIFFTAG_XRESOLUTION, (float)(sp->dpi));
		TIFFSetField(fo, TIFFTAG_YRESOLUTION, (float)(sp->dpi));
	} else {
		TIFFSetField(fo, TIFFTAG_XRESOLUTION, 300);
		TIFFSetField(fo, TIFFTAG_YRESOLUTION, 300);
	}
	if (comp) {
		if (e = i_CompressImage(sp,comp))
			goto error;
		TIFFSetField(fo, TIFFTAG_PHOTOMETRIC,
				(sp->d.rw.invert?PHOTOMETRIC_MINISBLACK:PHOTOMETRIC_MINISWHITE));
		if (e = writerawtiff(sp, fo))
			goto error;
	} else {
		TIFFSetField(fo, TIFFTAG_PHOTOMETRIC,PHOTOMETRIC_MINISWHITE);
		if (e = writeuntiff(sp, fo))
			goto error;
	}
error:
	if (fo) TIFFClose(fo);
	return e;
}
