
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "illum.h"

#include "c-lib.h"

#include <sys/types.h>
#include <sys/stat.h>

#include "rafutil.h"
#include "misc/prompt.h"

#include "ucs.h"
#include "utf.h"
#include "smemory.h"

#include "font.h"
#include "scroll.h"

#include "Document.h"
#include "MainWindow.h"
#include "Page.h"
#include "SelectionManager.h"
#include "ListDialog.h"
#include "ViewCanvas.h"
#include "ImageMode.h"
#include "OOCMode.h"

#include "TreeMode.h"

#include "RTEMode.h"
#include "OOCPage.h"

#include "text/text.h"

static void
disposeAuxHierarchy(iEntityPtr entity)
{
	iEntityPtr child;
	
	EntityAux *aux;

	// could have a non-traversing 
	// system if we kept a linked list
	for(child = i_GetChildEntity(entity); child; child = i_GetNextEntity(child)){
		disposeAuxHierarchy(child);
	}
	
	aux = getControllerMaybe(entity);
	if(aux) {
		removeController(entity);
		delete aux;
	}
}


/*
// written by David
*/

iDAFSError 
i_WriteEntityWBackup(const char *filename, 
					 const iEntityPtr entity, 
					 iDAFSWriteMode mode)
{
	FILE *fileptr;
	char backup[512];
	char overwrite = 0;
	iDAFSError e = DAFSOK;
	
#if 0
	if (fileptr = fopen(filename, "r")) {  /* file already exists */
		if (fclose( fileptr )) 
			e = WriteError;
		else {
			strcpy(backup, filename);
			strcat(backup, "~");
			if (rename(filename, backup))
				e = WriteError;
			overwrite = 1;
		}
	}
#endif
	if (fileptr = fopen(filename,"r")) { // file already exists
		if (fclose(fileptr))
			e = WriteError;
		overwrite = 1;
	}
	strcpy(backup,filename);
	strcat(backup,"~");
	if (!e){
		if (!(e = i_WriteEntity(backup, entity, mode))){
			if (overwrite)
				e = unlink(filename);
		}
		if (!e)
			if (rename(backup,filename))
				e = WriteError;
	}
	return e;
}

static void
hideModes(Document *doc)
{
	int i;

	for(i = 0; i < NUM_MODE_TYPES; i++){
		if(doc->modes[i]){
			doc->modes[i]->hide();
		}
	}
}

static void
showMode(Document *doc, ModeType vt)
{
	Mode *mode;
	aString correct_page;

	if(vt == vtOOC)
		correct_page = doc->last_ooc_page;

	mode = doc->modes[vt];
	
	if(mode->current_page_ref == 0) 
		mode->start();
	else 
		mode->show();

	// must do this first so selectPage works
	doc->current_mode=vt;

	if(vt == vtOOC)
		doc->selectPage(correct_page, doc->last_ooc_index);
}

static iDAFSError 
myUCode2Dafs(iUCode *text, iEntityPtr *doc_ptr, int swap)
{
	iEntityPtr doc;
	iEntityPtr line;
	iEntityPtr new_line;
	int newline_count = 0;
	iUCode u_newline = '\n';
	
	iUCode *s, *ubuffer=0;

	if(swap){
//		u_newline = i_SwapS(u_newline);
		for(s = text; *s; s++)
			*s = i_SwapS(*s);
	}
	
	doc = i_NewEntity(iDocumentT, 0);
	for(s = text; *s; *s ? s++ : s) {
		if(*s == u_newline){
			for(; *s; s++)
				if(*s == u_newline)
					newline_count++;
				else
					break;
			// at least 1 guaranteed
			new_line = i_NewEntity(iNewLineT, doc);
			i_SetShort(new_line, newline_count, 255);
			newline_count = 0;
			s--; // so it can be gobbled by the for loop
		} else {
			int n;
			const iUCode *beginning;

			if(newline_count){
				// there was a newline at the end of the last line
				new_line = i_NewEntity(iNewLineT, doc);
				i_SetShort(new_line, newline_count, 255);
				newline_count = 0;
			}
			line = i_NewEntity(iLineT, doc);

			// there is at least 1 character
			for(beginning = s; *s && *s != u_newline;s++) 
				;
			n = s-beginning;
			if (!(ubuffer = (iUCode *) smalloc((n+1)*sizeof(iUCode)))) 
				return MallocFailError;
			ucsncpy(ubuffer, beginning, n);
			ubuffer[n] = 0;
			i_SetUText(line, ubuffer,0);
			smfree(ubuffer); ubuffer = 0;
			// dont do an s-- because that would cause
			// an infinite loop on the last letter
			if(*s == u_newline)
				newline_count++;
		}
	}
	if(newline_count) {
		new_line = i_NewEntity(iNewLineT, doc);
		i_SetShort(new_line, newline_count, 255);
	}

	*doc_ptr = doc;

	return DAFSOK;
}

Document::Document(MainWindow *input_mw)
	: dirty(0),
	  root_entity(0),
	  mw(input_mw),
	  save_type(saveASCII),
	  current_mode(vtIMAGE),
	  original_mod_time(0)
{
	int i;
	for(i = 0; i < NUM_MODE_TYPES; i++)
		this->modes[i] = 0;
	this->last_ooc_index = 0;
}

Document::~Document()
{
	this->purgeModes();
	
	// assume there is one Document
	if(this->root_entity) {
		disposeAuxHierarchy(this->root_entity);
		i_DisposeEntity(this->root_entity);
		this->root_entity = 0;
	}
}


void
Document::purgeModes()
{
	for(int i = 0; i < NUM_MODE_TYPES; i++){
		if(this->modes[i]) {
			delete this->modes[i];
			this->modes[i] = 0; 
		}
	}
}


void
Document::purgeInactiveModes()
{
	for(int i = 0; i < NUM_MODE_TYPES; i++){
		if(this->modes[i] && i != current_mode) {
			this->modes[i]->purge();
			this->modes[i] = 0; 
		}
	}
}


void
Document::selectPage(const aString& pagename, int index)
{
	Mode *mode = this->getCurrentMode();
	DLList *l = &mode->page_list;
	Ref r;
	Page *page;
	
	for(r = dllistFirst(l); r; dllistNext(l, &r)){
		page = (Page *) dllistContents(l, r);
		if(!page->should_delete && page->name == pagename) {
			mode->setCurrentPage(r);

			if(index > 0){
				OOCPage *current_page = (OOCPage *) mode->getCurrentPage();
				int sub_page = index / (current_page->num_rows_page * current_page->num_cols);

				if(sub_page < current_page->num_pages)
					current_page->setPage(sub_page);
				else 
					current_page->setPage(current_page->num_pages-1);
			}
			break;
		}
	}
}


int
Document::initDoc(iEntityPtr entity, ModeType mode)
{
	int OK=1;
	
	this->root_entity = entity;
	
	//already done in closeFile?!?
	this->purgeModes();
	
	this->mw->changeMode(mode);
	
	if(this->modes[mode]) {
		this->clearDirty();
		this->mw->standardTitle();
		OK = 1;
	} else
		OK = 0;
	
	return OK;
}

/*
// bogus -- confusing mixture of int and iDAFSError
*/
int
Document::newFile()
{
	iEntityPtr doc;
	iDAFSError OK;
	
	disableCallbacks();
	doc = i_NewEntity(iDocumentT, 0);


	enableCallbacks();
	
	if (doc) {
		this->closeFile();
		this->save_type = saveASCII;
		OK=this->initDoc(doc, vtRTE);

		
		// there is no file
		this->file_name="";
		// the current_directory is unchanged

		// bogus we should do set standard title 
		this->mw->standardTitle();

		this->original_mod_time = 0;
	}   else
		OK=DAFSOK;
	
	return OK;
}

void
Document::setSaveType(SaveType save)
{
	this->save_type=save;
}

time_t
Document::calculateModTime()
{
	time_t mod_time;
	struct stat buf;
	int rv;
	
	rv = stat(this->file_name.string(), &buf);
	if(rv != 0) {
		mod_time = 0;
	} else {
		mod_time = buf.st_mtime;
	}
	
	return mod_time;
}
static int fileExists(const char *fn)
{
	FILE *f = fopen(fn, "r");
	if (f != NULL) {
		fclose(f);
		return TRUE;
	} else 
		return FALSE;
}


int
Document::saveFile()
{
	iDAFSError error_code;
	int return_code = 0;
	int accessible;
	int originally_on_disk;
	int modified_on_disk;
	int mod_time;
	int do_save = FALSE;
	FILE *fp;
	aString short_name = file_name.filename();

	if(this->dirtyP() == FALSE)
		return TRUE;
	
	originally_on_disk = this->original_mod_time != 0;
	
	fp = fopen(file_name.string(), "a");
	if(fp) {
		accessible = TRUE;
		fclose(fp);
	} else {
		accessible = FALSE;
	}
	
	mod_time = this->calculateModTime();
	
	modified_on_disk = (originally_on_disk && mod_time == 0 ||
		mod_time > this->original_mod_time) &&
				fileExists(file_name.string())==TRUE;
	
	if(accessible == FALSE) {
		aString foo;
		foo = aString("I couldn't save the file ") + short_name;
		
		uiWarnUser(foo.string(),
				   "OK");
		return_code = 0;
	} else {
		if(originally_on_disk == FALSE){
			do_save = TRUE;
		} else {
			if(modified_on_disk) {
				aString foo;
				int answer;
				foo = aString("File ") + short_name + " changed on disk. Overwrite?";
				
				answer = uiAskUser2(foo.string(),
									"Yes",
									"No",
									2);
				do_save = answer == 1;
			} else {
				do_save = TRUE;
			}
		}
		
		if(do_save && (file_name != "")) {
			if(i_ImageExt(file_name.string())) {
				aString foo;

				foo = "Use save as to save this image file in DAFSB format.";
				uiWarnUser(foo.string(),
						 "OK");
				return_code = 0;
			} else if(this->save_type==saveASCII) {
				char *text;
				FILE *foo;



				foo = fopen(file_name.string(), "w");
				
				text = i_AllocASCIIText(this->root_entity,0);
				if (text) {
					fwrite(text, 1, strlen(text), foo);
					i_FreeASCIIText(text);
				}
				fclose(foo);

				
				return_code = 1;
			} else if (this->save_type==saveUnicode) {
				iUCode buffer[4];
				FILE *foo;
				iUCode *us;

				
				foo = fopen(file_name.string(), "w");

				buffer[0]=0xfeff; /* write out byte swapping info */
				fwrite(buffer, sizeof(iUCode), 1, foo);

				us = i_AllocUText(this->root_entity);
				if (us) {
					fwrite(us,sizeof(iUCode), ucslen(us), foo);
					smfree(us);
				}
				fclose(foo);


				
				return_code = 1;
			} else {


				error_code = i_WriteEntityWBackup(file_name.string(),
												  this->root_entity,
												  DAFSB);


				if(error_code == DAFSOK) {
					return_code = 1;
				} else {
					aString foo;
					foo = aString("I couldn't save the file ") + short_name;
					
					uiWarnUser(foo.string(),
							   "OK");
					return_code = 0;
				}
			}
		}
	}
	
	if(return_code == 1) {
		this->clearDirty();
		this->original_mod_time = this->calculateModTime();
	}
    
	return return_code;
	// bogus check error conditions
}

void
Document::closeFile()
{
	this->mw->setModeSelection(0,0);
	this->getSM()->_clear();

	if(this->root_entity) {
		for (int i=0; i<NUM_MODE_TYPES; i++)
			if (this->modes[i]) {
				Ref r=this->modes[i]->current_page_ref;
				if (r) {
					Page *page=(Page *)dllistContents(&this->modes[i]->page_list,r);
					page->unmanage();
				}
			}
		disableCallbacks();
		
		// should be able to reorder these
		// nbut doing the dispose entity wasnt working and
		// does a lot of unnecessary list fiddling
		this->purgeModes();
		// ack -- necessary because TextMode fills 
		// it back in
		this->getSM()->_clear();
		
		disposeAuxHierarchy(this->root_entity);
		i_DisposeEntity(this->root_entity);
		
		this->root_entity = 0;
		enableCallbacks();
	}
	
	this->original_mod_time = 0;
}


int
Document::openFile(const FileName& fn, ModeType mode)
{
	enum LocalErrors { 
		leOK, 
		leFILE_NOT_FOUND,
		leLIMIT_EXCEEDED,
		leIMAGE_READ_FAIL, 
		leUTF_CONVERSION_FAILED,
		leDAFS_CONVERSION_FAILED,
		leDAFS_READ_FAIL,
	} local_error = leOK;
	iDAFSError error_code = DAFSOK;
	iEntityPtr entity = 0;
	iImagePtr image;
	aString ext;
	int OK = 1;
	int new_file = FALSE;
	int size;
	
	this->getSM()->clear();
	if(local_error == leOK) {
		this->would_be_savetype=saveDAFSB;
		ext = fn.extension();
		if(i_ImageExt(fn.string())) {
			error_code = i_ReadImage(fn.string(), &image);
			if (error_code == MallocFailError)
				halt("Out of memory\n");
			if(error_code == DAFSOK) {
				iBox box;
				disableCallbacks();
				entity = i_NewEntity(iDocumentT, 0);
				enableCallbacks();
			
				i_SetImage(entity, image);
				box.x = 0;
				box.y = 0;
				// mag doesn't count here
				box.w = i_GetImageWidth(image);
				box.h = i_GetImageHeight(image);
				i_SetBox(entity, box);
				mode=vtIMAGE; /* force it to be in image mode */
			} else if (!fileExists(fn.string())) {
					local_error = leFILE_NOT_FOUND;
			} else {
				local_error = leIMAGE_READ_FAIL;
			}
		} else {
			int is_dafs;
			is_dafs = dafsFileP(fn.string());
			if(is_dafs){
				disableCallbacks();
				error_code = i_ReadEntity(fn.string(), &entity);
				if (error_code == MallocFailError)
					halt("Out of memory\n");
				if(error_code != DAFSOK)
					local_error = leDAFS_READ_FAIL;
				enableCallbacks();
			} else {

				// read the whole file in
				char *s;
				iUCode *us;

	
				size = fileSizeInBytes(fn.string());
				if(size >= 80000)
					local_error = leLIMIT_EXCEEDED;
			
				if(local_error == leOK) {
					s = readAsciiFile(fn.string());
					disableCallbacks();
			
					new_file = !s;

					if (new_file) {
						// this means no file was present

						// bogus, should give user warning...
						s=(char *) smalloc(2);
						if(s == 0) 
							halt("ran out of memory.\n");
						s[0]='\0';
						s[1]='\0';
					}
					us=(iUCode *)s;
			
					iDAFSError e;
					// first character in unicode file is swap character
					int is_ucode;

					is_ucode =  *us==0xFFFE || *us==0xFEFF;
					if (is_ucode) {
						e = myUCode2Dafs(us+1,&entity,(*us==0xFFFE));
						if(e != DAFSOK)
							local_error = leDAFS_CONVERSION_FAILED;
						else {
							this->would_be_savetype=saveUnicode;
						}
					} else {
						// it is ascii/utf so convert it
						us = (iUCode *) smalloc((strlen(s) + 1) * sizeof(iUCode));
						int retval = utf2ucs(us, s);
						if(retval == -1){
							local_error = leUTF_CONVERSION_FAILED;
						} else {
#if 1
							// Im sick of bugs!
							e = myUCode2Dafs(us, &entity, 0);
#else
							// no longer inserts a newline
							e = i_UCode2Dafs(us, &entity, 0);
#endif
							if(e != DAFSOK)
								local_error = leDAFS_CONVERSION_FAILED;
							else {
								this->would_be_savetype=saveASCII;
							}
						}
						smfree(us);
					}

					enableCallbacks();
					smfree(s);
					if(local_error == leOK) {
						// is there a better way to set this -- so that
						// it jiggers the menu from here?
                        /* force it to be in text mode */

						mode=(mode==vtTREE)?vtTREE:vtRTE; 
					}
				}
			}
		}
	}	
	OK = 0;
	struct stat buf;
	aString error_message;
	aString short_name = fn.filename();

	switch(local_error){
	case leOK:
		OK = 1;
		break;
	case leFILE_NOT_FOUND:
		error_message = aString("Failed to read file ") + short_name 
			+ ". Could not open file.";
		break;
	case leLIMIT_EXCEEDED:
		error_message = aString("Internal limit exceeded for file ") + short_name 
			+ ". Try a smaller file.";
		break;
	case leUTF_CONVERSION_FAILED:
		error_message = aString("Failed attempt to read file ") + short_name 
			+ " as an ASCII/UTF file. "
			+ "Perhaps it is a binary file?";
		break;
	case leIMAGE_READ_FAIL:
		error_message = aString("Failed to read file ") + short_name 
			+ " as an Image file.";
		break;
	case leDAFS_READ_FAIL:
		error_message = aString("Failed to read file ") + short_name 
			+ " as an DAFS file.";
		break;
    case leDAFS_CONVERSION_FAILED:
		error_message = aString("Failed to convert text file ") + short_name 
			+ " into internal format.";
		break;
	}	

	if(error_message != ""){
		uiWarnUser(error_message.string(),
				   "OK");
	}
	
	if (OK) {
		this->original_mod_time = this->calculateModTime();
		this->closeFile();
		this->save_type=this->would_be_savetype;
		this->file_name = fn;
		// I think that directory should reset if you do a next or prev
		if (stat(fn.directory().string(),&buf))
			this->current_directory = getenv("PWD");
		else
			this->current_directory = fn.directory();
		if (entity) OK=this->initDoc(entity,mode);
		this->mw->refresh_dialogs();
	}
	
	this->mw->setSensitive();
	
	return OK;
}

Mode *
Document::getCurrentMode()
{
	return this->modes[current_mode];
}

void
Document::addNewEntity(iEntityPtr new_entity, int do_callback)
{
	int i;
	
	this->setDirty();
	for(i = 0; i < NUM_MODE_TYPES; i++)
		if(this->modes[i])
			this->modes[i]->addNewEntity(new_entity);
	
	if(do_callback)
		i_DoCallBack(mRecognize, new_entity);
}

void Document::setDirty()
{
	if(dirty == 0){
		dirty = 1; 
		mw->standardTitle();
	}
}

void Document::clearDirty()
{ 
	if(dirty) {
		dirty = 0; 
		mw->standardTitle();
	}
}

int 
Document::dirtyP() 
{ 
	return dirty; 
}

/*
// how do modes know to show the page that has the 
// selected entity
// 
// here it is
*/
void
Document::changeMode(ModeType vt)
{

	if(vt != this->current_mode || !this->modes[vt]) {
		switch(vt) {
		case vtIMAGE:
			this->goImageMode();
			this->mw->centerOnSelected();
			break;
		case vtOOC:
			this->goOOCMode();
			break;
		case vtFLAG:
			this->goFlagMode();
			this->mw->centerOnSelected();
			break;
		case vtTREE:
			this->goTreeMode();
			this->mw->centerOnSelected();
			break;

		case vtRTE:
			this->goRTEMode();
			this->mw->centerOnSelected();
			break;
		}

		this->purgeInactiveModes();
	}
}

void
Document::goOOCMode()
{
	hideModes(this);
	
	if(this->modes[vtOOC] == 0) {
		this->modes[vtOOC] = new OOCMode(this->mw, this->root_entity,iGlyphT,FALSE);
		this->modes[vtOOC]->setParentWidget(this->mw->scroll_window);
	}
	showMode(this, vtOOC);
}

void
Document::goFlagMode()
{
	iEntityPtr sole_entity;
	hideModes(this);
	
	if(this->modes[vtFLAG] == 0) {
		sole_entity = this->getSM()->getSoleItem();
		sole_entity = sole_entity && entityFlaggedP(sole_entity) ? sole_entity : 0;
		if(sole_entity == 0)
			this->getSM()->clear();
		this->modes[vtFLAG] = new OOCMode(this->mw, this->root_entity,0,TRUE);
		this->modes[vtFLAG]->setParentWidget(this->mw->scroll_window);
	}
	showMode(this, vtFLAG);
}

void
Document::goImageMode()
{
	hideModes(this);
	
	if(this->modes[vtIMAGE] == 0) {
		this->modes[vtIMAGE] = new ImageMode(this->mw, this->root_entity);
		this->modes[vtIMAGE]->setParentWidget(this->mw->scroll_window);
	}
	showMode(this, vtIMAGE);
}



void
Document::goRTEMode()
{

	hideModes(this);

	if(this->modes[vtRTE] == 0) {
		this->modes[vtRTE] = new RTEMode(this->mw, this->root_entity);
		this->modes[vtRTE]->setParentWidget(this->mw->scroll_window);
	}
	showMode(this,vtRTE);
}


void
Document::goTreeMode()
{
        hideModes(this);
        if(this->modes[vtTREE] == 0) {
                this->modes[vtTREE] = new TreeMode(this->mw, this->root_entity);
                this->modes[vtTREE]->setParentWidget(this->mw->scroll_window);
        }
        showMode(this, vtTREE);
}

void
Document::configure()
{
	int i;
	int4 x, y;
	Page *current_page = 0;
	Mode *mode = this->getCurrentMode();

	if (!mode || (current_mode != vtOOC && current_mode != vtFLAG))
		return;
	current_page = mode->getCurrentPage();
	if(current_page)
		getScrollPositionCenter(current_page->mode->scrolled_window, &x, &y);
	
	for(i = 0; i < NUM_MODE_TYPES; i++){
		if(this->modes[i]) {
			this->modes[i]->configure();
			this->modes[i]->hide();
		}
	}
	
	mode->show();
	if(current_page){
		ViewCanvas *vc = current_page->getCanvas();
		vc->scrollToPosition(current_page->mode->scrolled_window, x, y);
	}
}

SelectionManager *
Document::getSM()
{
	return this->mw->sm;
}
