
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"

#include <stdlib.h>
#include <stdio.h>

#include <Xm/CutPaste.h>
#include <Xm/DrawingA.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>
#include <Xm/RowColumn.h>

#include "rafutil.h"
#include "ucs.h"
#include "utf.h"
#include "smemory.h"
#include "uString.h"

#include "misc/Unicode.h"
#include "misc/XArg.h"
#include "misc/color.h"

#include "keyboard/KeyMap.h"

#include "dafs.h"
#include "Line.h"
#include "LineList.h"
#include "misc/oscroll.h"

#include "UniEditor.h"
#include "misc/Canvas.h"
#include "utf.h"

// forward declarations 
void uniInputCallback(Widget widget, UniEditor *uni, 
					  XmDrawingAreaCallbackStruct *cbs);
void uniExposeCallback(Widget widget, UniEditor *uni, 
					   XmDrawingAreaCallbackStruct *cbs);
void uniResizeCallback(Widget widget, UniEditor *uni, 
					   XmDrawingAreaCallbackStruct *cbs);

void pressButtonHandler(Widget widget, UniEditor *uni, XEvent *event);
void motionButtonHandler(Widget widget, UniEditor *uni, XEvent *event);
void releaseButtonHandler(Widget widget, UniEditor *uni, XEvent *event);
void pressKeyHandler(Widget widget, UniEditor *uni, XEvent *event);



UniEditor::UniEditor(Canvas *canvas, int input_scroll_flag)
  : CanvasClient(canvas),
	begin_line(0), begin_col(0),
	end_line(0), end_col(0),
	justification_direction(dLEFT),
	cursor_mode(cmBEFORE),
	selection_type(stNONE),
	scroll_flag(input_scroll_flag),
	canvas_width(0), canvas_height(0),
	cursor(0),
	entity_mode(FALSE),
	root_entity(0)
{
	// initialize data 
	overstrike = 0;
	selecting = FALSE;
	x_margin = y_margin = 4; // default margins
	goal_col = -1; // negative value signals don't use

#if 0
	// start up drawing area widget
	da = XtVaCreateWidget("drawing_area", xmDrawingAreaWidgetClass, parent, 
			      NULL);
	XtAddEventHandler(da, Button1MotionMask, False, 
					  (XtEventHandler) motionButtonHandler, (XtPointer) this);
#if 0
	XtAddCallback(da, XmNinputCallback, 
				  (XtCallbackProc) uniInputCallback, (XtPointer) this);
#endif

	XtAddEventHandler(da, ButtonPressMask, FALSE,
					  (XtEventHandler) pressButtonHandler, this);
	XtAddEventHandler(da, ButtonReleaseMask, FALSE,
					  (XtEventHandler) releaseButtonHandler, this);
	XtAddEventHandler(da, ButtonMotionMask, FALSE,
					  (XtEventHandler) motionButtonHandler, this);
	XtAddEventHandler(da, KeyPressMask, FALSE, 
					  (XtEventHandler) pressKeyHandler, this); 
	
	XtAddCallback(da, XmNexposeCallback, 
				  (XtCallbackProc) uniExposeCallback, (XtPointer) this);

	if(scroll_flag == FALSE)
		XtAddCallback(da, XmNresizeCallback, 
					  (XtCallbackProc) uniResizeCallback, (XtPointer) this);
#endif
	
	Widget widget = this->getWidget();
	XGCValues gcv;
	gcv.foreground = BlackPixelOfScreen(XtScreen(widget));
	XtVaGetValues(widget, XmNbackground, &background_color, NULL);
	gcv.background = background_color;

	gc = XCreateGC(XtDisplay(widget), RootWindowOfScreen(XtScreen(widget)), 
		       GCForeground | GCBackground, &gcv);
	XtVaSetValues(widget, XmNuserData, gc, NULL);	

	Widget da = this->getWidget();
	Widget parent = XtParent(da);
	gcv.foreground = getColorByName(parent, "red");
	cursor_gc = XCreateGC(XtDisplay(widget), RootWindowOfScreen(XtScreen(widget)),
			      GCForeground | GCBackground, &gcv);
	
}

UniEditor::~UniEditor()
{
  Widget canvas_widget = this->getWidget();
  if(cursor)
    XFreeCursor(XtDisplay(canvas_widget), cursor);
  XFreeGC(XtDisplay(canvas_widget), gc);
  XFreeGC(XtDisplay(canvas_widget), cursor_gc);
}

void
UniEditor::setString(const iUCode *string)
{
  Widget canvas_widget = this->getWidget();
	lines.deleteAll();

	entity_mode = FALSE;

	iUCode *temp = (iUCode *) smalloc((ucslen(string)+1) * sizeof(iUCode));
	ucscpy(temp, string);

	int offset = 0;
	iUCode *end, *start = temp;
	Line *line = lines.line[0];
	while(end = (iUCode *) ucschr(start, '\n')) {
	  *end = 0;
	  line->setText(start);
	  line->calculateLength(XtDisplay(canvas_widget));
	  lines.newLine();
	  line = lines.line[lines.size-1];
	  offset += (int) (end-start);
	  start = end + 1;
	}
	line->setText(start);	
	line->calculateLength(XtDisplay(canvas_widget));

	smfree(temp);

	this->setCursor(0, 0);

	this->conditionallyResize();
}

iUCode *
UniEditor::getString(void)
{
	int size = 0;
	int i;
	for(i = 0; i < lines.size; i++)
		size += lines.line[i]->textLength() + 1;

	iUCode *end, *string = (iUCode *) smalloc(sizeof(iUCode) * size);
	end = string;
	for(i = 0; i < lines.size; i++) {
	        Line *line = lines.line[i];
		ucsncpy(end, line->getText(), line->textLength());
		end += line->textLength();
		*end = (iUCode) '\n';
		end++;
	}
	end[-1] = 0;

	return string;
}

void
UniEditor::setEntity(iEntityPtr root)
{
  this->root_entity = root;
  Widget canvas_widget = this->getWidget();
  lines.deleteAll();
  lines.parseEntity(root);
  lines.parseContext();
  for(int i = 0; i < lines.size; i++)
    lines.line[i]->calculateLength(XtDisplay(canvas_widget));


  entity_mode = TRUE;

  // interfered with the version 
  // in illum starting off with a translation of the selection in
  // the other view
  // other versions of UniEditor can explicitly do this if they like
  //this->setCursor(0, 0);

  this->conditionallyResize();
}

/* 
// questions
// what are real lengths?
// when do we add a newline
//
// bogus -- rewrite after reading blammo selection
*/
iUCode *
UniEditor::getString(int first_line, int first_col,
					 int second_line, int second_col)
{
	int size = 0;
	int len;
	Line *line;
	int i;

	for(i = first_line; i <= second_line; i++)
		size += lines.line[i]->textLength() + 1;
	/* it will be a little over, because we might be in the middle
	   of the first or last lines */

	iUCode *end, *string = (iUCode *) smalloc(sizeof(iUCode) * size);

	end = string;

	if(first_line == second_line) {
		line = lines.line[first_line];
		len = second_col - first_col;
		ucsncpy(end, line->getText() + first_col, len);
		end += len;
#if 0
		*end = (iUCode) '\n';
		end++;
#endif
	} else {
		// first line
		line = lines.line[first_line];
		len = line->textLength() - first_col;
		ucsncpy(end, line->getText() + first_col, len);
		end += len;
		*end = (iUCode) '\n';
		end++;
	
		for(i = first_line + 1; i < second_line; i++) {
			line = lines.line[i];
			ucsncpy(end, line->getText(), line->textLength());
			end += line->textLength();
			*end = (iUCode) '\n';
			end++;
		}

		// second line
		line = lines.line[second_line];
		len = second_col;
		ucsncpy(end, line->getText(), len);
		end += len;
		end++;
	}
	
	end[-1] = 0;

	return string;
}

void
UniEditor::lookAtKeyMap()
{
  KeyMap *keymap = this->getKM();
  int direction;
  Widget canvas_widget = this->getWidget();

  iUCode letter;
  keymap->getInfo(&direction, &letter);
  direction = NORMAL;
  this->justification_direction = direction == NORMAL ? dLEFT : dRIGHT;
  this->justification_direction = dLEFT;
  measureUnicode(XtDisplay(canvas_widget), &letter, 1, &this->char_width, &this->char_height, &ascent);
  this->char_height = 25; // maximum height of all glyphs in our unicode fontsets
}


unsigned 
UniEditor::pos2line(int y)
{
	int line_pos;
	
	line_pos = (y - y_margin)/this->char_height;
	if(line_pos < 0)
		line_pos = 0;
	if(line_pos >= lines.size)
		line_pos = lines.size-1;

	return line_pos;
}

unsigned
UniEditor::pos2base(int click_x, int line_pos)
{
  Widget canvas_widget = this->getWidget();
  int glyph_width;
  if(justification_direction == dLEFT)
    glyph_width = x_margin;
  else
    glyph_width = this->canvas_width - x_margin - this->char_width;

  int base = 0;
  for(int i = 0; i < lines.line[line_pos]->contextLength() + 1; i++) {
    if((justification_direction == dLEFT  && glyph_width > click_x) ||
       (justification_direction == dRIGHT && glyph_width < click_x)) {
      base--;
      break;
    }

    if(i >= lines.line[line_pos]->contextLength())
      break;
    
    iUCode glyph = lines.line[line_pos]->getContext(i);
    if(uniSpacing(glyph)) {
      int w, h, a;
      measureUnicode(XtDisplay(canvas_widget), &glyph, 1, &w, &h, &a);
      if(justification_direction == dLEFT)
	glyph_width += w;
      else
	glyph_width -= w;
    }
    base++;
  }

  if(base < 0) base = 0;
  base = lines.line[line_pos]->context2base(base);
  return base;
}

void
UniEditor::resize(XmDrawingAreaCallbackStruct *cbs)
{
  Widget canvas_widget = this->getWidget();

	Dimension width, height;
	
	XtVaGetValues(canvas_widget, XmNwidth, &width, XmNheight, &height, NULL);
	this->canvas_width  = width;
	this->canvas_height = height;
}

void  
UniEditor::exposeCallback(XmDrawingAreaCallbackStruct *cbs)
{
  Widget canvas_widget = this->getWidget();

	XEvent *event = cbs->event;

	if(!cursor) {
	  cursor = XCreateFontCursor(XtDisplay(canvas_widget), XC_xterm);
	  XDefineCursor(XtDisplay(canvas_widget), XtWindow(canvas_widget), cursor);
	}

	int y1 = event->xexpose.y;
	int y2 = event->xexpose.y+event->xexpose.height;
	int initial = pos2line(y1),
	    final   = pos2line(y2);
	drawBorder();
	drawRange(initial, final);
}

void
UniEditor::processKeyPress(XEvent *event)
{
  Widget canvas_widget = this->getWidget();

	char key_value;
	KeySym ks;
	int left_line, left_col, right_line, right_col;
	int fail = FALSE;
		
	int ctrl_on = event->xkey.state & ControlMask,
	    alt_on  = event->xkey.state & Mod1Mask;
	
	int state = event->xkey.state;

	// bogus should look for shift key to make left, right
	// etc extend selections
	XLookupString(&event->xkey,
				  &key_value,
				  1,
				  &ks,
				  NULL);
		
	// blow off early if ks is a modifier or if selecting
	if(ks >= XK_Shift_L && ks <= XK_Hyper_R)
		return;

	int initial = 0, 
	    final  = this->lastLine();
	switch(ks){
	case XK_BackSpace:
		this->backspaceAtSelection(&initial, &final);
		break;	  
	case XK_Delete:
		this->deleteAtSelection(&initial, &final);
		break;
	case XK_Left: 
	  this->left(state);
	  break;
	case XK_Right:
	  this->right(state);
	  break;
	case XK_Insert:
		overstrike = !overstrike;
		break;

	case XK_End:
		if(ctrl_on) {
			this->setCursor(this->lastLine(), this->lastCol(this->lastLine()));
		} else {
			switch(this->getSelectionType()) {
			case stSELECTION:
				this->getSelection(&left_line, &left_col, &right_line, &right_col);
				break;
			case stCURSOR:
				this->getCursor(&right_line, &right_col);
				break;
			case stNONE:
				fail = TRUE;
				XBell(XtDisplay(canvas_widget), 0);
			}
			if(!fail) {
				this->setCursor(right_line, this->lastCol(right_line));
				initial = final = -1; // setCursor handles drawing
			}
		}
		break;
	case XK_Home:
		if(ctrl_on) {
			this->setCursor(0, 0);
		} else {
			switch(this->getSelectionType()) {
			case stSELECTION:
				this->getSelection(&left_line, &left_col, &right_line, &right_col);
				break;
			case stCURSOR:
				this->getCursor(&left_line, &left_col);
				break;
			case stNONE:
				fail = TRUE;
				XBell(XtDisplay(canvas_widget), 0);
			}
			if(!fail) {
				this->setCursor(left_line, 0);
			}
		}
		initial = final = -1; // setCursor handles drawing
		break;
	case XK_Up:
	  this->up(state);
	  return;
	  break;
	case XK_Down:
	  this->down(state);
	  return;
	  break;
	case XK_Return:
		this->insertAtSelection('\n', &initial, &final);
		break;
	default:
		iUCode char_value;
		// convert keysym to unicode
		KeyMap *km;
		km = this->getKM();
		if(alt_on)
			char_value = km->lookupControlValue(ks);
		else
			char_value = km->lookupValue(ks);
		this->insertAtSelection(char_value, &initial, &final);
	}

#if 0
	// set goal_col unless up or down (which return, not break from switch loop)
	if(selection_type == stCURSOR) {
	  int screen_col = lines.line[begin_line]->base2screen(begin_col);
	  goal_col = lines.line[begin_line]->renderLength(XtDisplay(canvas_widget), screen_col);
	}
#endif

	this->conditionallyResize();
	this->scrollIfNecessary();
	this->drawRange(initial, final);
}

int
UniEditor::pressKey(iUCode c)
{
  int initial = 0, 
	final  = this->lastLine();

  this->insertAtSelection(c, &initial, &final);

  this->conditionallyResize();
  this->scrollIfNecessary();
  this->drawRange(initial, final);
  return TRUE;
}

#if 0
void 
UniEditor::inputCallback(XmDrawingAreaCallbackStruct *cbs)
{
	XEvent *event = cbs->event;

	switch(event->xany.type) {
	case ButtonPress: 
		processButtonPress(event);
		break;
	case ButtonRelease:
		processButtonRelease(event);
		break;
	case KeyPress: 
		processKeyPress(event);
		break;
	case KeyRelease:
		break;
	}
}
#endif





void 
motionButtonHandler(Widget w, UniEditor *uni, XEvent *event)
{
  uni->moveButton(event->xmotion.state, event->xmotion.x, event->xmotion.y);
}

void 
pressButtonHandler(Widget w, UniEditor *uni, XEvent *event)
{
	uni->pressButton(event);
}

void 
releaseButtonHandler(Widget w, UniEditor *uni, XEvent *event)
{
	uni->releaseButton(event);
}

void 
pressKeyHandler(Widget w, UniEditor *uni, XEvent *event)
{
	uni->processKeyPress(event);
}

void
uniExposeCallback(Widget widget, UniEditor *uni, XmDrawingAreaCallbackStruct *cbs)
{
	uni->exposeCallback(cbs);
}

void
uniResizeCallback(Widget widget, UniEditor *uni, XmDrawingAreaCallbackStruct *cbs)
{
	uni->resize(cbs);
}

#if 0
void 
uniInputCallback(Widget widget, UniEditor *uni, XmDrawingAreaCallbackStruct *cbs)
{
	uni->inputCallback(cbs);
}
#endif

void
UniEditor::coordToPosition(int x, int y, int *line, int *column)
{
	*line = pos2line(y);
	*column = pos2base(x, *line);
}

void 
UniEditor::positionToCoord(int line, int column, cMode cm, int *x, int *y)
{
  Widget canvas_widget = this->getWidget();

	// bogus -- only gets y approximately right
	Line *l =  this->lines.line[line];
	*y = line * this->char_height;
	int context = l->base2context(column);
	*x = x_margin + l->renderLength(XtDisplay(canvas_widget),
					context + (cm == cmBEFORE ? 0 : 1));
}

void
UniEditor::selectionDrawRegion(int *first_line, int *last_line)
{
	int foo;
	switch(this->getSelectionType()) {
	case stNONE:
		*first_line = -1;
		*last_line = -1;
		break;
	case stCURSOR:
		this->getCursor(first_line, &foo);
		*last_line = *first_line;
		break;
	case stSELECTION:
		this->getSelection(first_line, &foo,
				   last_line, &foo);
		break;
	}
}

void
UniEditor::pressButton(XEvent *event)
{
  Widget canvas_widget = this->getWidget();

	int line;
	int column;

	switch(event->xbutton.button){
	case 1:
		xstart = event->xbutton.x;
		ystart = event->xbutton.y;
		selecting = TRUE;
		this->coordToPosition(xstart, ystart, &line, &column);
		if(this->setCursor(line, column)) {
		  if(xstart < x_margin && lines.line[line]->direction(column) == dRIGHT)
		    this->setCursorMode(cmAFTER);
		  else
		    this->setCursorMode(cmBEFORE);
		}
		break;
	case 2:
		int status;
		char buffer[1024];
			
		while((status = XmClipboardRetrieve(XtDisplay(canvas_widget), XtWindow(canvas_widget), "STRING",
						    buffer, sizeof buffer, NULL, NULL)) == ClipboardLocked);

#if 0
		if(status == ClipboardSuccess){
			iUCode u_buffer[1024];
			utf2ucs(u_buffer, buffer);

			int raw    = lines.line[line_on]->base2text(insertion),
				length = ucslen(u_buffer);
			//				lines.line[line_on]->move(raw, length);
			for(int i = 0; i < length; i++)
				lines[line][raw+i] = u_buffer[i];					

			drawAll();
		}
#endif
		break;
	}
}

void
UniEditor::moveButton(int state, int x, int y)
{
	int changed;

	if(selecting){
		// must have pressed button 1
#if 0
		int x = event->xmotion.x,
		    y = event->xmotion.y;
#endif

		int o_line;
		int o_column;
		int n_line;
		int n_column;
		this->coordToPosition(x, y, &n_line, &n_column);
		this->coordToPosition(this->xstart, this->ystart, &o_line, &o_column);
#if 0
		printf("%3d,%3d  %3d,%3d\n", 
			   o_line, o_column, 
			   n_line, n_column);
		printf("%3d,%3d  %3d,%3d\n", 
			   o_line, o_column, 
			   n_line, n_column);
#endif

		if(o_line < n_line) {
			changed = this->setSelection(o_line, o_column, n_line, n_column);
		} else if (o_line > n_line) {
			changed = this->setSelection(n_line, n_column, o_line, o_column);
		} else if (o_column < n_column) {
			changed = this->setSelection(o_line, o_column, n_line, n_column);
		} else if(o_column > n_column) {
			changed = this->setSelection(n_line, n_column, o_line, o_column);
		} else 
			changed = this->setCursor(o_line, o_column);
	}
	
	// the set functions do their own drawing
}


void 
UniEditor::releaseButton(XEvent *event)
{
	if(event->xbutton.button == 1){
		selecting = FALSE;
		
		// simply wrong because it wants to use line_on
#if 0
		// copy the data to the clipboard
		int status;
		unsigned long item_id = 0;
		XmString label = XmStringCreateSimple("Unicode Widget");
		while((status = XmClipboardStartCopy(XtDisplay(canvas_widget), XtWindow(canvas_widget), label, CurrentTime, 
											 NULL, NULL, &item_id)) == ClipboardLocked);
		XmStringFree(label);
				
		unsigned start = lines.line[line_on]->base2text(start_pos),
			stop  = lines.line[line_on]->base2text(stop_pos + 1);
		stop += uniSpacing(stop);

		iUCode *u_buffer = (iUCode *) smalloc((stop-start+1)*sizeof(iUCode));
		ucsncpy(u_buffer, lines.line[line_on]->getText()+start, stop-start);
		u_buffer[stop-start] = 0;
		char *buffer = (char *) smalloc((stop-start+1)*4);  // max of 4 bytes per unicode char in utf
		ucs2utf(buffer, u_buffer);
				
		while((status = XmClipboardCopy(XtDisplay(canvas_widget), XtWindow(canvas_widget), item_id, "STRING",
										buffer, strlen(buffer)+1, 0, NULL)) == ClipboardLocked);

		smfree(buffer);
		smfree(u_buffer);
				
		while((status = XmClipboardEndCopy(XtDisplay(canvas_widget), XtWindow(canvas_widget), item_id)) == ClipboardLocked);
#endif
	}
}

iLBox 
UniEditor::getDesiredSize(void)
{
	int w, h;
	int i;
	int line_w;
	iLBox value;

	this->lookAtKeyMap();

 	w = 0;
	h = this->lines.size * this->char_height;
	for(i = 0; i < lines.size; i++) {
		line_w = lines.line[i]->renderLength();
		w = MAX(line_w, w);
	}
	w += 2 * x_margin;
	h += 2 * y_margin;

	value.x = value.y = 0;
	value.w = w;
	value.h = h;

	return value;
}
 
int
UniEditor::resizeNecessaryP()
{
	iLBox size;
	size = this->getDesiredSize();
	return size.w != this->canvas_width || size.h != this->canvas_height;
}

void
UniEditor::conditionallyResize()
{
  if(resizeNecessaryP()) {
	iLBox size;
	size = this->getDesiredSize();
	this->setSize(size.w, size.h);
  }
}

void
UniEditor::setSize(int w, int h)
{
  Widget canvas_widget = this->getWidget();


	// bogus -- maybe should make minimum size in accordance
	// with scrolled window size
  if(this->scroll_flag) {
	iBox visible;
	if(XtIsRealized(canvas_widget))
	  visible = visibleRectangleAuto(XtParent(XtParent(canvas_widget)));
	else 
	  visible.w = visible.h = 0;
	
	w = MAX(w, visible.w);
	h = MAX(h, visible.h);
  } else {
	w = MAX(w, 10);
	h = MAX(h, 10);
  }

  if(this->scroll_flag) {
	// otherwise the height of the window is the height of the
	// window and we can't make it happen by wishing
	// only if scott overrides are we really going to resize
	this->canvas_height = h;
	this->canvas_width = w;
	XArg args;
	args.setArg(XmNwidth, w);
	args.setArg(XmNheight, h);
	args.setValues(canvas_widget);
  }
}

int
UniEditor::isVisible(int x, int y)
{
  Widget canvas_widget = this->getWidget();
  if(XtIsRealized(canvas_widget) == FALSE)
	return FALSE;

	if(this->scroll_flag) {
		Widget sw = XtParent(XtParent(canvas_widget));
		return visiblePositionPAuto(sw, x, y);	
	} else {
		return TRUE;
	}
}

void
UniEditor::calculateVisible(int *first, int *last)
{
  Widget canvas_widget = this->getWidget();

	int foo;

	if(this->scroll_flag) {
		Widget sw = XtParent(XtParent(canvas_widget));
		iBox vr;

		if(XtIsRealized(canvas_widget))
		  vr = visibleRectangleAuto(sw);
		else
		  vr.w = vr.h = vr.x = vr.y = 0;
		this->coordToPosition(vr.x, vr.y, first, &foo);
		this->coordToPosition(vr.x + vr.w, vr.y + vr.h, last, &foo);
	} else {
		XArg args;
		Dimension h;
		args.setArg(XmNheight, &h);
		args.getValues(canvas_widget);
		*first = 0;
		this->coordToPosition(0, h, last, &foo);
	}
}

void
UniEditor::centerOnCoord(int x, int y)
{
  Widget canvas_widget = this->getWidget();
	Widget sw = XtParent(XtParent(canvas_widget));
	scrollToPositionAuto(sw, x, y);
}


void
UniEditor::up(int state)
{
  Widget canvas_widget = this->getWidget();

  int left_line; int left_col;
  int right_line; int right_col;
  int fail = FALSE;

  switch(this->getSelectionType()) {
  case stSELECTION:
	this->getSelection(&left_line, &left_col, &right_line, &right_col);
	break;
  case stCURSOR:
	this->getCursor(&left_line, &left_col);
	break;
  case stNONE:
	fail = TRUE;
	XBell(XtDisplay(canvas_widget), 0);
  }
  if(!fail) {
	if(left_line > 0) {
	  Line *line = lines.line[left_line];
	  int screen_col, prev_col, x_pos;

	  screen_col = line->base2screen(left_col);
	  if(goal_col < 0)
	    x_pos = line->renderLength(XtDisplay(canvas_widget), screen_col);
	  else
	    x_pos = goal_col;
	  prev_col = pos2base(x_pos+x_margin, left_line - 1);

	  this->setCursor(left_line - 1, prev_col);
	  this->setCursorMode(cmBEFORE);
	} else 
	  this->setCursor(left_line, left_col);
  }

  this->conditionallyResize();
}

void
UniEditor::down(int state)
{
  Widget canvas_widget = this->getWidget();

  int left_line; int left_col;
  int right_line; int right_col;
  int fail = FALSE;

  switch(this->getSelectionType()) {
  case stSELECTION:
	this->getSelection(&left_line, &left_col, &right_line, &right_col);
	break;
  case stCURSOR:
	this->getCursor(&right_line, &right_col);
	break;
  case stNONE:
	fail = TRUE;
	XBell(XtDisplay(canvas_widget), 0);
  }
  if(!fail) {
	if(right_line < this->lastLine()) {
	  Line *line = lines.line[right_line];
	  int screen_col, next_col, x_pos;

	  screen_col = line->base2screen(right_col);
	  if(goal_col < 0)
	    x_pos = line->renderLength(XtDisplay(canvas_widget), screen_col);
	  else
	    x_pos = goal_col;
	  next_col = pos2base(x_pos+x_margin, right_line + 1);

	  this->setCursor(right_line + 1, next_col);
	  this->setCursorMode(cmBEFORE);
	} else 
	  this->setCursor(right_line, right_col);
  }

  this->conditionallyResize();
}

void
UniEditor::left(int state)
{
  Widget canvas_widget = this->getWidget();

  int left_line; int left_col;
  int right_line; int right_col;
  int fail = FALSE;


  switch(this->getSelectionType()) {
  case stSELECTION:
	this->getSelection(&left_line, &left_col, &right_line, &right_col);
	break;
  case stCURSOR:
	this->getCursor(&left_line, &left_col);
	break;
  case stNONE:
	fail = TRUE;
	XBell(XtDisplay(canvas_widget), 0);
  }
  if(!fail) 
        this->moveCursor(dLEFT);

  this->conditionallyResize();
}


void
UniEditor::right(int state)
{
  Widget canvas_widget = this->getWidget();

  int left_line; int left_col;
  int right_line; int right_col;
  int fail = FALSE;

  switch(this->getSelectionType()) {
  case stSELECTION:
	this->getSelection(&left_line, &left_col, &right_line, &right_col);
	break;
  case stCURSOR:
	this->getCursor(&right_line, &right_col);
	break;
  case stNONE:
	fail = TRUE;
	XBell(XtDisplay(canvas_widget), 0);
  }
  if(!fail) 
	this->moveCursor(dRIGHT);

  this->conditionallyResize();
}

Widget
UniEditor::getWidget()
{
  return this->getCanvas()->canvas_widget;
}

void
UniEditor::drawRegionNow(iLBox region)
{
  int x1;
  int x2;
  int y1;
  int y2;
  x1 = region.x;
  x2 = region.x + region.w;
  y1 = region.y;
  y2 = region.y + region.h;
  int first_line;
  int last_line;
  int foo;

  this->coordToPosition(x1, y1, &first_line, &foo);
  this->coordToPosition(x2, y2, &last_line, &foo);
  this->drawRange(first_line, last_line);
}



void 
UniEditor::home(int state) 
{
  Widget canvas_widget = this->getWidget();
  int fail = FALSE;
  int ctrl_on = state & ControlMask,
	alt_on  = state & Mod1Mask;
  int left_line, left_col, right_line, right_col;

  if(ctrl_on) {
	this->setCursor(0, 0);
  } else {
	switch(this->getSelectionType()) {
	case stSELECTION:
	  this->getSelection(&left_line, &left_col, &right_line, &right_col);
	  break;
	case stCURSOR:
	  this->getCursor(&left_line, &left_col);
	  break;
	case stNONE:
	  fail = TRUE;
	  XBell(XtDisplay(canvas_widget), 0);
	}
	if(!fail) {
	  this->setCursor(left_line, 0);
	}
  }

  this->scrollIfNecessary();
}

void 
UniEditor::end(int state)
{
  Widget canvas_widget = this->getWidget();
  int fail = FALSE;
  int initial = 0, 
	final  = this->lastLine();
  int ctrl_on = state & ControlMask,
	alt_on  = state & Mod1Mask;
  int left_line, left_col, right_line, right_col;

  if(ctrl_on) {
	this->setCursor(this->lastLine(), this->lastCol(this->lastLine()));
  } else {
	switch(this->getSelectionType()) {
	case stSELECTION:
	  this->getSelection(&left_line, &left_col, &right_line, &right_col);
	  break;
	case stCURSOR:
	  this->getCursor(&right_line, &right_col);
	  break;
	case stNONE:
	  fail = TRUE;
	  XBell(XtDisplay(canvas_widget), 0);
	}
	if(!fail) {
	  this->setCursor(right_line, this->lastCol(right_line));
	}
  }

  this->scrollIfNecessary();
}

void 
UniEditor::myReturn()
{
  Widget canvas_widget = this->getWidget();

  int initial = 0, 
	final  = this->lastLine();

  this->insertAtSelection('\n', &initial, &final);

  this->conditionallyResize();
  this->scrollIfNecessary();
  this->drawRange(initial, final);
}

void 
UniEditor::backspace()
{
  Widget canvas_widget = this->getWidget();

  int initial = 0, 
	final  = this->lastLine();

  this->backspaceAtSelection(&initial, &final);

  this->conditionallyResize();
  this->scrollIfNecessary();
  this->drawRange(initial, final);
}

void 
UniEditor::myDelete()
{
  Widget canvas_widget = this->getWidget();

  int initial = 0, 
	final  = this->lastLine();

  this->deleteAtSelection(&initial, &final);

  this->conditionallyResize();
  this->scrollIfNecessary();
  this->drawRange(initial, final);
}

void 
UniEditor::insert()
{
#if 0
  this->overstrike = !this->overstrike;
#endif
}


void
mypr(char *s)
{
  while(*s) {
	if(*s == '\n')
	  printf("\\n");
	else
	  printf("%c", *s);
	s++;
  }
  printf("\n");
}

void
UniEditor::pr()
{
  printf("begin UniEditor.\n");
  this->prSelf();

  for(int i = 0; i < this->lines.size; i++) {
    Line *line = this->lines.line[i];
	printf("\tbegin line %d.\n", i);
	line->pr();
	printf("\tend line %d.\n", i);
  }
  printf("end UniEditor.\n");
  printf("\n\n");
}

void
UniEditor::prSelf()
{
  printf("begin UniEditor self.\n");
  iUCode *s = this->getString();
  char *ascii = (char *) smalloc(utfMaxSize(s) * sizeof(char));
  ucs2utf(ascii, s);
  mypr(ascii);
  smfree(ascii);
  smfree(s);
  printf("end UniEditor self.\n");
}

void
UniEditor::compare(UniEditor *other)
{
  int min_lines = MIN(this->lines.size, other->lines.size);
  int i;
  int success;
  if(this->lines.size != other->lines.size){
	printf("mismatch in the number of lines original:%d copy:%d\n",
		   this->lines.size,
		   other->lines.size);
  }
  for(i = 0; i < min_lines; i++){
	Line *my_line = this->lines.line[i];
	Line *other_line = other->lines.line[i];
	printf("line %d\n", i);
	success = my_line->compare(other_line);
  }
}
