
#include "RTEView.h"
#include "PackedRectangle.h"
#include "RTECursor.h"


// --------------------------- PackedRectangle methods -------------------------

PackedRectangle::PackedRectangle()
{
	primary_writing_direction = DEFAULT_PRIM_READ_ORDER;
	secondary_writing_direction = DEFAULT_SCND_READ_ORDER;
	spacing = DEFAULT_LINE_SPACING;
}

PackedRectangle::PackedRectangle(Direction pwd, Direction swd, Coord space)
{
	if(pwd == swd || pwd == opposite(swd)) halt("illegal combination of directions");
	if(space < 0) halt("space < 0");
	this->primary_writing_direction = pwd;
	this->secondary_writing_direction = swd;
	this->spacing = space;
}

PackedRectangle::PackedRectangle(Coord x, Coord y, Coord w, Coord h,
				 Direction pwd, Direction swd, Coord space) :
	ParentRectangle(x, y, w, h)
{
	if(pwd == swd || pwd == opposite(swd)) halt("illegal combination of directions");
	if(space < 0) halt("space < 0");
	this->primary_writing_direction = pwd;
	this->secondary_writing_direction = swd;
	this->spacing = space;
}

void PackedRectangle::debugPrint(char *s)
{
	debugf("Contents of PackedRectangle %s:\n", s);
	this->h_interval_list->debugPrint("horizontal Intervals");
	this->v_interval_list->debugPrint("vertical Intervals");
}

Coord
PackedRectangle::getSpacing(void)
{
	Coord result=0;
	if(this->spacing)
	  result = this->spacing;
	else switch(secondary_writing_direction) {
	case LeftToRight:
	case RightToLeft:
	  result = this->w;
	  break;
	case TopToBottom:
	case BottomToTop:
	  result = this->h;
	  break;
	default:
	  halt("invalid direction\n");
	  break;
	}
	return result;
}

void PackedRectangle::changeChildSize(Coord new_w, Coord new_h, Rectangle *child)
{
	// Change child's size while keeping it aligned with my line spacing.

	if(this->spacing != 0) {
		assert(this->isAlignedWithLineSpacing(child));
	
		// keep leading point in packing order at same location
		switch(primary_writing_direction) {
			case TopToBottom:
				break;
			case BottomToTop:
				child->y -= new_h - child->h;
				break;
			case LeftToRight:
				break;
			case RightToLeft:
				child->x -= new_w - child->w;
				break;
			default:
				halt("No secondary packing order\n");
		}
		// align with line spacing but keep inside me
		// heinous - don't change child->x without re-registering with parent

		switch(secondary_writing_direction){
			case TopToBottom:
			case BottomToTop:
				child->y += child->h - new_h;
				child->y = MAX(0, child->y);
				break;
			case LeftToRight:
			case RightToLeft:
				child->x -= (new_w - child->w)/2;
				if(this->w < this->spacing)
					child->x -= (this->spacing - this->w)/2;
				child->x = MAX(0, child->x);
				break;
			default:
				halt("No secondary packing order\n");
		}
	}
	ParentRectangle::changeChildSize(new_w, new_h, child);
	assert(this->isAlignedWithLineSpacing(child));
}

int PackedRectangle::inOrder(ChildRectangle *r1, ChildRectangle *r2)
{
	// Find the corner point of each Rectangle which is "first"
	// in both the primary and secondary reading directions.

	if(r1 == r2) 
		return EQUAL;
	else {
		const Direction pwd = this->primary_writing_direction,
	                	swd = this->secondary_writing_direction;
	
		Coord x1=r1->getX(), y1=r1->getY(), x2=r2->getX(), y2=r2->getY();
	
		if(pwd == RightToLeft || swd == RightToLeft) {
			x1 += r1->getW()-1;
			x2 += r2->getW()-1;
		}
	
		if(pwd == BottomToTop || swd == BottomToTop) {
			y1 += r1->getH()-1;
			y2 += r2->getH()-1;
		}
	
		if(x1 == x2 && y1 == y2) {
			// Rectangles occupy same position in packing order; 
			// smaller one comes first
			const int a1 = r1->area(), a2 = r2->area();
			if(a1 < a2)
				return LESS_THAN;
			else if(a1 == a2)
				return EQUAL;
			else
				return GREATER_THAN;
		} else
			return this->inOrder(x1, y1, x2, y2);
	}
}

int PackedRectangle::inOrder(Coord x1, Coord y1, Coord x2, Coord y2)
{
	// Measure the distance of each point along the packing order

	int pcoord1, pcoord2, scoord1, scoord2, result;

	Coord spacing;
	if(this->spacing)
	  spacing = this->spacing;
	else switch(secondary_writing_direction) {
	case LeftToRight:
	case RightToLeft:
	  spacing = this->w;
	  break;
	case TopToBottom:
	case BottomToTop:
	  spacing = this->h;
	  break;
	default:
	  break;
	}

	switch (this->secondary_writing_direction) {
		case LeftToRight:
			pcoord1 = x1/spacing;  pcoord2 = x2/spacing;
			break;
		case RightToLeft:
			pcoord1 = (this->w - x1)/spacing;  pcoord2 = (this->w - x2)/spacing;
			break;
		case TopToBottom:
			pcoord1 = y1/spacing;  pcoord2 = y2/spacing;
			break;
		case BottomToTop:
			pcoord1 = (this->h - y1)/spacing;  pcoord2 = (this->h - y2)/spacing;
			break;
		case RTENone:
			halt("Secondary reading direction is RTENone\n");
			break;
	}

	if(pcoord1 < pcoord2)
		result = LESS_THAN;
	else if(pcoord1 > pcoord2)
		result = GREATER_THAN;
	else {
		switch (this->primary_writing_direction) {
			case LeftToRight:
				scoord1 = x1;  scoord2 = x2;
				break;
			case RightToLeft:
				scoord1 = this->w - x1;  scoord2 = this->w - x2;
				break;
			case TopToBottom:
				scoord1 = y1;  scoord2 = y2;
				break;
			case BottomToTop:
				scoord1 = this->h - y1;  scoord2 = this->h - y2;
				break;
			case RTENone:
				halt("Primary reading direction is RTENone\n");
				break;
		}
		if(scoord1 < scoord2)
			result = LESS_THAN;
		else if(scoord1 > scoord2)
			result = GREATER_THAN;
		else
			result = EQUAL;
	}
	return result;
}

void PackedRectangle::deleteChild(ChildRectangle *c, int squeeze, int erase)
{
	assert(this->isParentOf(c));
	if(squeeze == UNSPEC) {
	  if(this->lastChild() != c)
	    squeeze = !c->allowOverlap();
	  else
	    squeeze = TRUE;
	}
	assert(squeeze == TRUE || squeeze == FALSE);

	// ChildRectangle *nc = this->nextChild(c, FALSE),
	               // *pc = this->prevChild(c, FALSE);
	// const int shrink = (!nc || !nc->positionIsValid()) || (!pc || !pc->positionIsValid());

	if(erase) c->requestErase();
	if(c->position_valid) {
		c->setPositionNotValid();
		this->freeArea(c, c, squeeze);
	}

	#ifndef NDEBUG
	ElementSet *my_children = this->getChildrenInRectangle(this);
	assert(!my_children->contains(c));
	delete my_children; my_children = 0;
	#endif
	
	this->element_list->removeItemNumber(this->element_list->locationOfItem(c));
	c->parent = NULL;
	if(squeeze /* && shrink */)
		if(!this->shrinkToFitChildren()) {
			// debugf("shrinkToFitChildren failed\n");
			halt("shrinkToFitChildren failed\n");
		}
}

ChildRectangle *
PackedRectangle::firstChild(void)
{
  if(this->element_list->getSize())
    return this->element_list->getItemNumber(0);
  else
    return NULL;
}

ChildRectangle *
PackedRectangle::lastChild(void)
{
  const int size = this->element_list->getSize();

  if(size)
    return this->element_list->getItemNumber(size-1);
  else
    return NULL;
}


ChildRectangle * PackedRectangle::nextChild(ChildRectangle *c, int wrap)
{
	assert(this->isParentOf(c));
	ChildRectangle *result=0;
	const int loc = this->element_list->locationOfItem(c);
	if(loc + 1 < this->element_list->getSize())
		result = this->element_list->getItemNumber(loc + 1);
	if(result && !wrap && this->differentLine(c, result))
		result = 0;
	return result;
}

ChildRectangle * PackedRectangle::prevChild(ChildRectangle *c, int wrap)
{
	assert(this->isParentOf(c));
	ChildRectangle *result=0;
	const int loc = this->element_list->locationOfItem(c);
	if(loc > 0)
		result = this->element_list->getItemNumber(loc - 1);
	if(result && !wrap && this->differentLine(c, result))
		result = 0;
	return result;
}


void PackedRectangle::positionOnTopOfChild(ChildRectangle *c, Rectangle *r)
{
	// Position r so that it occupies the same place in the packing
	// order as c  

        switch(this->primary_writing_direction) {
                case LeftToRight:
                        r->x = c->x;
			r->y = c->y - (r->h - c->h);
                        break;
                case RightToLeft:
                        r->x = c->x - (r->w - c->w);
			r->y = c->y - (r->h - c->h);
                        break;
                case TopToBottom:
			r->x = c->x - (r->x - c->x)/2;
                        r->y = c->y;
                        break;
                case BottomToTop:
			r->x = c->x - (r->x - c->x)/2;
                        r->y = c->y - (r->h - c->h);
                        break;
                case RTENone:
                        halt("primary writing direction is RTENone\n");
                        break;
        }
        this->alignWithLineSpacing(r);
}

void PackedRectangle::positionBeforeChild(ChildRectangle *c, Rectangle *r)
{
	// Note:  may set r to an illegal position if c occupies the
	// first position in a line.

        switch (this->primary_writing_direction) {
                case RightToLeft:
                        r->x = c->x + c->w;
			r->y = c->y - (r->h - c->h);
                        break;
                case LeftToRight:
                        r->x = c->x - r->w;
			r->y = c->y - (r->h - c->h);
                        break;
                case TopToBottom:
			r->x = c->x - (r->x - c->x)/2;
                        r->y = c->y - r->h;
                        break;
                case BottomToTop:
			r->x = c->x - (r->x - c->x)/2;
                        r->y = c->y + c->h;
                        break;
                case RTENone:
                        halt("primary writing direction is RTENone");
                        break;
        }
	this->alignWithLineSpacing(r);
}

int PackedRectangle::positionAfterChild(ChildRectangle *c, Rectangle *r)
{
  // Set x and y of r so that it is positioned immediately after c
  // in my reading order.  If c is at the end of a line, return the
  // first position in the next line.  If c is in the last possible position
  // in the packing order, we fail.
  
  if(c->allowOverlap())
    positionOnTopOfChild(c, r);
  else {
    switch (this->primary_writing_direction) {
    case RightToLeft:
      r->x = c->x - r->w;
      r->y = c->y - (r->h - c->h);
      break;
    case LeftToRight:
      r->x = c->x + c->w;
      r->y = c->y - (r->h - c->h);
      break;
    case BottomToTop:
      r->x = c->x - (r->x - c->x)/2;
      r->y = c->y - r->h;
      break;
    case TopToBottom:
      r->x = c->x - (r->x - c->x)/2;
      r->y = c->y + c->h;
      break;
    case RTENone:
      halt("primary writing direction is RTENone");
      break;
    }
    this->alignWithLineSpacing(r);
  }	
  
  return TRUE;
}

static int lmlt(int x, int k)
{
	// return the largest multiple of k which is <= x
	assert(k > 0 && x >= 0);
	return x - (x % k);
}

void PackedRectangle::positionFirst(Rectangle *r)
{
	// Set x and y of r so that it occupies the first position in 
	// my reading order.

	// Note: there is no guarantee that r will lie inside the
	// packed rectangle after this routine because r could be
	// too big to fit inside me.

	assert(primary_writing_direction != secondary_writing_direction &&
	       primary_writing_direction != opposite(secondary_writing_direction));

	Coord spacing;
	if(this->spacing)
	  spacing = this->spacing;
	else switch(secondary_writing_direction) {
	case LeftToRight:
	case RightToLeft:
	  spacing = this->w;
	  break;
	case TopToBottom:
	case BottomToTop:
	  spacing = this->h;
	  break;
	default:
	  break;
	}

	switch (primary_writing_direction) {
		case LeftToRight: r->x = 0;                           break;
		case RightToLeft: r->x = MAX(this->w - r->w, 0);      break;
		case TopToBottom: r->y = 0;                           break;
		case BottomToTop: r->y = MAX(this->h - r->h, 0);      break;
		case RTENone: halt("primary writing direction is RTENone"); break;
	}

	switch (secondary_writing_direction) {
		case LeftToRight: r->x = 0;                                   break;
		case RightToLeft: r->x = lmlt(this->w, this->spacing) - r->w; break;
		case TopToBottom: r->y = this->spacing - r->h;                break;
		case BottomToTop: r->y = lmlt(this->h, this->spacing) - r->h; break;
		case RTENone: halt("secondary writing direction is RTENone");       break;
	}
	this->alignWithLineSpacing(r);
}

void PackedRectangle::moveForward(Coord *px, Coord *py, Coord move_amt)
{
	// Move (*px, *py) forward by move_amt in the primary writing direction

	switch(primary_writing_direction) {
		case TopToBottom: *py += move_amt; break;
		case BottomToTop: *py -= move_amt; break;
		case LeftToRight: *px += move_amt; break;
		case RightToLeft: *px -= move_amt; break;
		case RTENone: halt("primary writing direction is RTENone"); break;
	}
}

void PackedRectangle::moveForward(Rectangle *r, Coord move_amt)
{
	this->moveForward(&r->x, &r->y, move_amt);
}

int PackedRectangle::positionAtStartOfNextLine(Coord *px, Coord *py, Coord rw, Coord rh)
{
	// Set *px and *py so that (*px, *py) occupies the first position in 
	// the line following the one which it currently occupies.
	// Assume that (x,y) is currently alligned with my line spacing.

	// NOTE: calculated position may be illegal - may go outside parent

	assert(primary_writing_direction != secondary_writing_direction &&
	       primary_writing_direction != opposite(secondary_writing_direction));

	Coord spacing;
	if(this->spacing)
	  spacing = this->spacing;
	else switch(secondary_writing_direction) {
	case LeftToRight:
	case RightToLeft:
	  spacing = this->w;
	  break;
	case TopToBottom:
	case BottomToTop:
	  spacing = this->h;
	  break;
	default:
	  break;
	}

	assert(this->isAlignedWithLineSpacing(*px, *py, rw, rh));
	switch (secondary_writing_direction) {
		case TopToBottom:
			*py += spacing;
			break;
		case BottomToTop:
			*py -= spacing;
			break;
		case LeftToRight:
			*px += spacing;
			break;
		case RightToLeft:
			*px -= spacing;
			break;
		case RTENone:
			halt("secondary writing direction is RTENone\n");
			break;
	}
	switch (primary_writing_direction) {
		case LeftToRight:
			*px = 0;
			break;
		case RightToLeft:
			*px = this->w - rw;
			break;
		case BottomToTop:
			*py = this->h - rh;
			break;
		case TopToBottom:
			*py = 0;
			break;
		case RTENone:
			halt("primary writing direction is none\n");
			break;
	}
	assert(this->isAlignedWithLineSpacing(*px, *py, rw, rh));
	return TRUE;
}

int PackedRectangle::positionAtStartOfNextLine(Rectangle *r)
{
	// Set r->x and r->y so that r occupies the first position
	// (in packing order) in the line following the one which
	// it currently occupies.

	// NOTE: calculated position may be illegal - may go outside parent

	return this->positionAtStartOfNextLine(&r->x, &r->y, r->w, r->h);
}

int PackedRectangle::isAlignedWithLineSpacing(Coord x, Coord y, Coord w, Coord h)
{
	Coord spacing;
	if(this->spacing)
	  spacing = this->spacing;
	else switch(secondary_writing_direction) {
	case LeftToRight:
	case RightToLeft:
	  spacing = this->w;
	  break;
	case TopToBottom:
	case BottomToTop:
	  spacing = this->h;
	  break;
	default:
	  break;
	}

	int result = FALSE;
	switch (secondary_writing_direction) {
		case TopToBottom:
		case BottomToTop:
			if(this->h < spacing)
				result = TRUE;
			else
				result = ((y + h) % spacing == 0);
			break;
		case LeftToRight:
			if(this->w <spacing)
				result = TRUE;
			else
				result = ((x - 1 + (w + spacing)/2) % spacing == 0);
			break;
		case RightToLeft:
			if(this->w < spacing)
				result = TRUE;
			else
				result = ((x + (w + spacing)/2) % spacing == 0);
			break;
		case RTENone:
			halt("secondary writing direction is RTENone\n");
			result = FALSE;
			break;
	}
	return result;
}

int PackedRectangle::isAlignedWithLineSpacing(Rectangle *r)
{
	return this->isAlignedWithLineSpacing(r->x, r->y, r->w, r->h);
}

void PackedRectangle::alignWithLineSpacing(Rectangle *r)
{
 	// When the secondary writing direction is vertical (i.e. lines of
        // text are horizontal) child Rectangle is positioned as far down
        // as possible.  When secondary writing direction is horizontal (i.e.
        // lines of text are vertical) child Rectangle is centered.

	Coord space;
	int center;

	Coord spacing;
	if(this->spacing)
	  spacing = this->spacing;
	else switch(secondary_writing_direction) {
	case LeftToRight:
	case RightToLeft:
	  spacing = this->w;
	  break;
	case TopToBottom:
	case BottomToTop:
	  spacing = this->h;
	  break;
	default:
	  break;
	}

	switch (secondary_writing_direction) {
		case TopToBottom:
		case BottomToTop:
			space = MIN(spacing, this->h);
			r->y = nearestPositiveMultiple(r->y + r->h, space) - r->h;
			break;
		case LeftToRight:
			space = MIN(spacing, this->w);
			center = (spacing <= this->w);
			center = FALSE;
			// bogus -- I don't think centering is necessarily
			// a good idea and it isn't doing what I want now
			// revisit this later
			if(center) {
				r->x = nearestPositiveMultiple(r->x + r->w, space) - 
			       	(space + r->w)/2 + (space + r->w) % 2;
			} else {
				// left align
				r->x = 0;
			}
			break;
		case RightToLeft:
			space = MIN(spacing, this->w);
			center = (spacing <= this->w);
			center = FALSE;
			// bogus -- I don't think centering is necessarily
			// a good idea and it isn't doing what I want now
			// revisit this later
			if(center) {
				r->x = nearestPositiveMultiple(r->x + r->w, space) - 
			       	(space + r->w)/2;
			} else {
				// right align
				r->x = MAX(0, this->w - r->w);

			}
			break;
		case RTENone:
			halt("secondary writing direction is RTENone\n");
			break;
	}
	assert(this->isAlignedWithLineSpacing(r));
}

void PackedRectangle::alignWithEdge(Rectangle *r, RTECursor *cur)
{
	// Align a rectangular cursor with either my leading or trailing
	// edge in the primary packing direction.

	Coord cx, cy;
	cur->getPosition(&cx, &cy);
	switch (this->primary_writing_direction) {
		case LeftToRight:  case RightToLeft:  case RTENone:
			if(cx > r->x + r->w/2)
				cx = r->x + r->w;
			else
				cx = r->x;
			switch (this->secondary_writing_direction) {
				case TopToBottom:
					cy = r->y + r->h - cur->getSize();
					break;
				case BottomToTop:  case RTENone:
					cy = r->y;
					break;
				default: 
					halt("illegal secondary writing direction");
					break;
			}
			break;
		case BottomToTop:  case TopToBottom:
			if(cy > r->y + r->h/2)
				cy = r->y + r->h;
			else
				cy = r->y;
			switch (this->secondary_writing_direction) {
				case LeftToRight:
					cx = r->x;
					break;
				case RightToLeft:
					cx = r->x + r->w - cur->getSize();
					break;
				default: 
					halt("illegal secondary writing direction");
					break;
			}
			break;
	}
	cur->moveTo(cx, cy, FALSE);
}

int PackedRectangle::insertFirst(ChildRectangle *new_c)
{
	assert(!this->isParentOf(new_c));
	int result;
	ElementList *vcl = this->getAllValidChildren();
	if(vcl->getSize() > 0) 
		result = insertBefore(vcl->getItemNumber(0), new_c);
	else {
		this->positionFirst(new_c);
		result = new_c->becomeChildOf(this);
	}
	return result;
}

int PackedRectangle::insertLast(ChildRectangle *new_c)
{
	assert(!this->isParentOf(new_c));
	int result;
	ElementList *vcl = this->getAllValidChildren();
	const int n = vcl->getSize();
	if(n > 0) 
		result = this->insertAfter(vcl->getItemNumber(n-1), new_c);
	else
		result = this->insertFirst(new_c);
	delete vcl;
	return result;
}

int PackedRectangle::insertAt(Coord px, Coord py, ChildRectangle *new_c, int allow_overlap)
{
	assert(!this->isParentOf(new_c));
	const int redraw = TRUE;
	new_c->x = px;  new_c->y = py;
	return new_c->becomeChildOf(this, redraw, allow_overlap);
}

void PackedRectangle::trimExtraSize(void)
{
	// to be implemented
}

int PackedRectangle::changeSizeTo(Coord new_w, Coord new_h, int redraw)
{
	// Note:  if we are attempting to change my size to something which
	// is too small to hold all my children (after repacking) I will grow
	// large enough to hold them.  This growth will occur according to my
	// canGrow members, and may exceed the specified new_w and new_h.

	int success=FALSE;

	if(new_w == this->w && new_h == this->h)
		success = TRUE; // nothing to do
	else {
		if(!this->needToRepack(new_w-this->getW(), new_h-this->getH()))
			success = this->ParentRectangle::changeSizeTo(new_w, new_h, redraw);
		else {
			ElementList *original_child_list = this->getAllValidChildren();
			success = this->ChildRectangle::changeSizeTo(new_w, new_h, FALSE);
			if(success) {
				const Coord old_w = this->getW(), old_h = this->getH();
				this->requestErase();
				this->deleteAllValidChildren();
				if(!this->repackChildren(original_child_list)) {
					success = FALSE;
					if(!this->ChildRectangle::changeSizeTo(old_w, old_h, FALSE))
						halt("PackedRectangle::changeSizeTo - unable to restore original size\n");
					else
						if(!this->repackChildren(original_child_list))
							halt("PackedRectangle::changeSizeTo - unable to restore children to original packing\n");
				}
			delete original_child_list;
			}
		}
	}
	return success;
}

int PackedRectangle::containsParents()
{
	int result=FALSE;
	for(int i=0; i<this->numChildren() && !result; i++) {
		if(this->getChildNumber(i)->numChildren() > 0)
			result = TRUE;
	}
	return result;
}

int PackedRectangle::needToRepack(Coord dw, Coord dh)
{
	// See if I need to repack my children if I change size by (dw, dh)

	int result=TRUE;
	if(this->numChildren() == 0)
		result = FALSE;
	else {
		switch(this->primary_writing_direction)
		{
  		case LeftToRight: 
			if((dw == 0 && !this->containsParents()) ||
			   (dh == 0 && this->getH() <= this->getSpacing()))
				result = FALSE;
      			break;
  		case RightToLeft:
			if((dw == 0 && !this->containsParents()) ||
			   (dh == 0 && this->getH() <= this->getSpacing()))
				result = FALSE;
      			break;
    		case TopToBottom:
    		case BottomToTop:
			if((dh == 0 && !this->containsParents()) ||
			   (dw == 0 && this->getH() <= this->getSpacing()))
				result = FALSE;
      			break;
    		default:
      			halt("primary writing direction not defined\n");
      			break;
		}
		Rectangle *r = new Rectangle(0, 0, this->getW()+dw, this->getH()+dh);
		for(int i=0; i<this->numChildren() && !result; i++)
			result = !r->contains(this->getChildNumber(i));
		delete r;
	}
	return result;
}

int PackedRectangle::repackChildren(ElementList *cl)
{
	int result = TRUE;
	if(!cl) {
		if(this->numChildren() > 0) {
			cl = this->getAllValidChildren();
			this->deleteAllValidChildren();
		}
	}
	if(cl) {
		for(int i=0; result && i<cl->numElements(); i++) {
			ChildRectangle *c = cl->getItemNumber(i);
	
			// See if child c needs to be repacked itself before being
			// reinserted into me.
			// heinous - we need a better criteria for need to repack
			const int need_to_repack = FALSE;
			//const int need_to_repack = (c->numChildren() > 0);
			if(need_to_repack){
				if(c->maximizeWidth())
					this->growChildToFill(c);
				else 
					this->minimizeChildSize(c);
				result = this->insertLast(c);
				if(result) {
					PackedRectangle *pc = (PackedRectangle *) c;
					if(!pc->repackChildren())
						halt("PackedRectangle::repackChildren failed to repack child of child\n");
				}
			} else {
				if(c->maximizeWidth())
					this->growChildToFill(c);
				result = this->insertLast(c);
			}
		}
	}
	return result;
}

int PackedRectangle::growChildToFill(ChildRectangle *c)
{
	Coord new_w, new_h;
	PackedRectangle *pc;
	switch (c->getPrimaryWritingDir())
	{
	case LeftToRight:
	case RightToLeft:
		pc = (PackedRectangle *) c;
		new_w = this->getW();
		new_h = MIN(pc->getSpacing(), pc->getH());
		break;
	case TopToBottom:
	case BottomToTop:
		pc = (PackedRectangle *) c;
		new_w = MIN(pc->getSpacing(), pc->getW());
		new_h = this->getH();
		break;
	case RTENone:
		return TRUE;  // do nothing
		break;
	}
	const int redraw=FALSE;
	return c->ChildRectangle::changeSizeTo(new_w, new_h, redraw);
}

int PackedRectangle::shrinkChildToFit(ChildRectangle *c)
{
	const Coord new_w = MIN(this->getW(), c->getW()),
	            new_h = MIN(this->getH(), c->getH());
	const int redraw=FALSE;
	return c->ChildRectangle::changeSizeTo(new_w, new_h, redraw);
}

int PackedRectangle::minimizeChildSize(ChildRectangle *c)
{
	const Coord min_w = 1, min_h = 1;
	const int redraw = FALSE;
	return c->ChildRectangle::changeSizeTo(min_w, min_h, redraw);
}

int PackedRectangle::insertBefore(ChildRectangle *c, ChildRectangle *new_c)
{
	const int allow_overlap = new_c->allowOverlap();
	return this->insertBefore(c, new_c, allow_overlap);
}

int PackedRectangle::insertAfter(ChildRectangle *c, ChildRectangle *new_c)
{
	const int allow_overlap = c->allowOverlap();
	return this->insertAfter(c, new_c, allow_overlap);
}

int PackedRectangle::insertBefore(ChildRectangle *c, ChildRectangle *new_c, int allow_overlap)
{
	// Insert c_new where c currently is, moving c and subsequent children
	// forward in the packing direction
	
	assert(this->isParentOf(c) && !this->isParentOf(new_c));
	int result = FALSE;
	if(c->allowOverlap()) {
		ChildRectangle *pc = this->prevChild(c);
		if(pc) {
			const int squeeze=FALSE, redraw=FALSE;
			this->deleteChild(c, squeeze, redraw);
			result = this->insertAfter(pc, new_c) && this->insertAfter(new_c, c);
		} else {
			this->positionOnTopOfChild(c, new_c);
			result = this->insertAt(new_c->x, new_c->y, new_c, allow_overlap);
        		if(!result) {
                		positionAtStartOfNextLine(&new_c->x, &new_c->y, new_c->w, new_c->h);
				result = this->insertAt(new_c->x, new_c->y, new_c, allow_overlap);
        		}
		}
		if(result) c->requestRedraw();
	} else {
		this->positionOnTopOfChild(c, new_c);
		result = this->insertAt(new_c->x, new_c->y, new_c, allow_overlap);
        	if(!result) {
                	positionAtStartOfNextLine(&new_c->x, &new_c->y, new_c->w, new_c->h);
			result = this->insertAt(new_c->x, new_c->y, new_c, allow_overlap);
        	}
	}
        return result;
}

int PackedRectangle::insertAfter(ChildRectangle *c, ChildRectangle *new_c, int allow_overlap)
{
	// Insert c_new after c, moving subsequent child Rectangles forward
	// if needed.

	assert(this->isParentOf(c) && !this->isParentOf(new_c));
	const int redraw = TRUE;
	this->positionAfterChild(c, new_c);
        int result = new_c->becomeChildOf(this, redraw, allow_overlap);
        if(!result) {
                positionAtStartOfNextLine(&new_c->x, &new_c->y, new_c->w, new_c->h);
                result = new_c->becomeChildOf(this, redraw, allow_overlap);
        }
        return result;
}

static Coord moveNeededToClear(ElementList *cl, Rectangle *r, Direction dir)
{
	// Calculate how far *c must be moved in the direction dir to 
	// have no intersection with *r

	Coord move_amt=0;
	int subtract_cursor_width=FALSE;
	ChildRectangle *c = cl->getItemNumber(0);

	#if 0
	if(c->allowOverlap()) {
		if(cl->numElements() == 1)
			subtract_cursor_width = TRUE;
		else
			c = cl->getItemNumber(1);
	}
	#endif

	switch (dir) {
		case RightToLeft:
			move_amt = (c->getX() + c->getW() - 1) - r->getX() + 1;
			if(subtract_cursor_width)
				move_amt -= c->getW();
			break;
		case LeftToRight:
			move_amt = (r->getX() + r->getW() - 1) - c->getX()  +  1;
			//if(move_amt < 0)
				//halt("moveNeededToClear calculated negative move amt\n");
			// if(subtract_cursor_width)
				// move_amt -= c->getW();
			break;
		case TopToBottom:
			move_amt = (r->getY() + r->getH() - 1) - c->getY()  +  1;
			if(subtract_cursor_width)
				move_amt -= c->getH();
			break;
		case BottomToTop:
			move_amt = (c->getY() + c->getH() - 1) - r->getY() + 1;
			if(subtract_cursor_width)
				move_amt -= c->getH();
			break;
		case RTENone:  
			halt("Direction is RTENone");
			break;
	}
	return move_amt;
}

#ifndef NDEBUG
static int isClear(ParentRectangle *p, Rectangle *r, ChildRectangle *cr)
{
	ElementSet *es = p->getChildrenInRectangle(r);
	es->removeElement(cr);
	int result = es->isEmpty();
	delete es;
	return result;
}
#endif

int PackedRectangle::allocateArea(Coord x1, Coord y1, Coord x2, Coord y2, 
								  ChildRectangle *cr)
{
	int success = FALSE;
	ChildRectangle *r = new ChildRectangle(x1, y1, x2-x1+1, y2-y1+1);

	if(r) {
		success = this->allocateArea(r, cr);
		delete r; r = 0;
	}
	return success;
}

int PackedRectangle::allocateArea(ChildRectangle *r, ChildRectangle *cr)
{
	// allocate the rectangular area r for the child Rectangle cr

	int success = FALSE;
	if(r->area() == 0) return TRUE;
	if(this->contains(r))  {
		ElementList *conflicts=0;
		if(!(success = ParentRectangle::allocateArea(r, cr, &conflicts))) {
			if(conflicts && !conflicts->isEmpty()) { 
				Coord move_amt = moveNeededToClear(conflicts, r, primary_writing_direction);
				Coord dx, dy;
				if(this->moveChildrenForward(conflicts, move_amt, &dx, &dy)) {
					assert(isClear(this, r, cr));
					success = r->moveBy(dx, dy);
					if(r != cr)
						success = success && cr->moveBy(dx, dy);
					success = success && registerAreaAllocation(r, cr);
				}
			}
			if(conflicts) { 
				delete conflicts; conflicts = 0; 
			}
		}
	} else {
		success = FALSE;
		Rectangle *ri = this->intersection(r);
		if(ri){
			ElementList *toMove = this->getListOfChildrenInRectangle(ri);
			if(toMove && toMove->contains(cr)) 
				toMove->removeItem(cr);
			if(toMove && !toMove->isEmpty()){
				Coord move_amt = moveNeededToClear(toMove, r, this->primary_writing_direction);
				Coord dx, dy;
				this->moveChildrenForward(toMove, move_amt, &dx, &dy);
				assert(isClear(this, r, cr));
				delete toMove;  toMove = 0;
			}
			delete ri; ri = 0;
		}
	}
	return success;
}

int PackedRectangle::moveChildForward(ChildRectangle *c, Coord move_amt, Coord *x_shift, Coord *y_shift)
{
	// Either move me forward by move_amt in the primary writing direction,
	// or to the beginning of the next line

	int success = TRUE;
	const int squeeze = FALSE;
	*x_shift = 0;
	*y_shift = 0;
	if(move_amt != 0) {
		success = FALSE;
		Coord cx = c->x, cy = c->y;
		this->moveForward(&cx, &cy, move_amt);
		success = c->moveTo(cx, cy, squeeze);
		if(success) {
			*x_shift = MAX(0, -cx);
			*y_shift = MAX(0, -cy);
		} else
			success = this->moveChildToStartOfNextLine(c);
	}
	return success;
}

int PackedRectangle::moveChildrenForward(ElementList *el, Coord move_amt, Coord *x_shift, Coord *y_shift)
{
	int success = TRUE;
	*x_shift = 0;
	*y_shift = 0;
	if(move_amt != 0) {
		for(int i=el->getSize()-1; i>=0 && success; i--) {
			Coord dx, dy;
			success = moveChildForward(el->getItemNumber(i), move_amt, &dx, &dy);
			*x_shift += dx;  *y_shift += dy;
		}
	}
	return success;
}

int PackedRectangle::moveChildToStartOfNextLine(ChildRectangle *c)
{
	const int squeeze = FALSE;
	Coord cx = c->x, cy = c->y;
	positionAtStartOfNextLine(&cx, &cy, c->w, c->h);
	return c->moveTo(cx, cy, squeeze);
}

int PackedRectangle::freeArea(Rectangle *r, ChildRectangle *cr, int squeeze=TRUE)
{
	// If squeeze is true, try to move one or more Rectangles to fill
	// the freed space.
	int success;
	if(r == (Rectangle *)cr)
		success =  this->ParentRectangle::freeArea(cr);
	else
		success =  this->ParentRectangle::freeArea(r, cr);
	if(success && squeeze) {
		// See if the elements following me in the packing order can move 
		// into the space I just freed.  This could cause a chain reaction
		// all the way to the end of the packed Rectangle.
		ChildRectangle *nc = this->nextChild(cr, TRUE),
		               *pc = this->prevChild(cr, TRUE);
		RTECursor *cur = NULL;

		if(nc && nc->getType() == CursorT) {
			cur = (RTECursor *) nc;
			nc = this->nextChild(nc, TRUE);
			if(nc)
				this->deleteChild(cur, FALSE, TRUE);
			else {
				nc = (ChildRectangle *) cur;
				cur = NULL;
			}
		}
                if(nc) {
			this->deleteChild(nc);
			if(pc)
				this->insertAfter(pc, nc);
			else
				this->insertFirst(nc);
		}
		if(cur) this->insertBefore(nc, cur);
			
	}
	return success;
}

Coord smgt(Coord, Coord);  // heinous - put this declaration into RTEMisc.h

int PackedRectangle::shrinkToFitChildren()
{
	int result;

	if(this->numChildren() == 0) {
		if(this->parent) {
			// heinous - assumes all parents are packed
			PackedRectangle *p = (PackedRectangle *) this->parent;
			if(this->maximizeWidth())
				result = p->growChildToFill(this);
			else 
				result = p->minimizeChildSize(this);
		}
	} else {
		Coord new_w, new_h;
		int redraw=FALSE;
		Rectangle *r_union = this->element_list->getItemNumber(0)->areaUnion(this->element_list);
		if(r_union) {
			switch(this->getPrimaryWritingDir())
			{
			case BottomToTop:
			case TopToBottom:
				if(this->spacing)
					new_w = smgt(r_union->getW(), this->spacing);
				else
					new_w = r_union->getW();
				if(this->maximizeWidth())
					new_h = this->getH();
				else
					new_h = r_union->getH();
				break;
			case LeftToRight:
			case RightToLeft:
				if(this->maximizeWidth())
					new_w = this->getW();
				else
					new_w = r_union->getW();
				if(this->spacing)
					new_h = smgt(r_union->getH(), this->spacing);
				else
					new_h = r_union->getH();
				break;
			default:
				halt("primary writing direction not defined\n");
				break;
			}
			delete r_union;
		}
		redraw = (new_h < this->getH() || new_w < this->getW());
		result = this->changeSizeTo(new_w, new_h, redraw);
	}
	return result;
}

#if 0
int PackedRectangle::shrinkToFitChildren()
{
	// Change my size in the secondary packing direction to the minimum
	// multiple of spacinig such that I contain all my children at their
	// current positions.

	Rectangle *r = new Rectangle(this->getX(), this->getY(), this->getW(), this->getH());
	ChildRectangle *cf = this->firstChild(), *cl = this->lastChild();
	int result = TRUE;

	const Coord spacing = this->getSpacing();

	for(int i=0; ; i++)
	{
		// see if reducing our size by (i+1) * spacing excludes cr
		switch (this->getSecondaryWritingDir())
		{
			case BottomToTop:
			case TopToBottom:
				r->setH(MAX(0, r->getH() - spacing));
				break;
			case LeftToRight:
			case RightToLeft:
				r->setW(MAX(0, r->getW() - spacing));
				break;
			default:
				halt("secondary writing direction not defined\n");
				break;
		}
		if(r->area() <= 0 || !r->contains(cf) || !r->contains(cl)) break;
	}
	if(i > 0) {
		Coord new_w, new_h;
		switch (this->getSecondaryWritingDir())
		{
			case BottomToTop:
			case TopToBottom:
				new_w = this->getW();
				new_h = MAX(0, this->getH() - i * spacing);
				break;
			case LeftToRight:
			case RightToLeft:
				new_w = MAX(0, this->getW() - i * spacing);
				new_h = this->getH();
				break;
			default:
				halt("secondary writing direction not defined\n");
				break;
		}
		int redraw = FALSE; // children erase and redraw themselves
		result = this->ChildRectangle::changeSizeTo(new_w, new_h, redraw);
	}
	return result;
}
#endif

int PackedRectangle::differentLine(ChildRectangle *c1, ChildRectangle *c2)
{
	int line1=0, line2=0;
	const int spacing = this->getSpacing();
	switch (this->secondary_writing_direction) {
		case TopToBottom:  case BottomToTop:
			line1 = c1->y / spacing,
			line2 = c2->y / spacing;
			break;
		case LeftToRight:  case RightToLeft:
			line1 = c1->x / spacing,
			line2 = c2->x / spacing;
			break;
		default:  halt("secondary writing direction is RTENone");
			break;
	}
	return (line1 != line2);
}

ChildRectangle *
PackedRectangle::firstLeafChild(ChildRectangle *child) {
  if(child->numChildren()) {
    PackedRectangle *parent = (PackedRectangle *) child;
    return parent->firstLeafChild(parent->firstChild());
  }
  return child;
}

ChildRectangle *
PackedRectangle::lastLeafChild(ChildRectangle *child) {
  if(child->numChildren()) {
    PackedRectangle *parent = (PackedRectangle *) child;
    return parent->lastLeafChild(parent->lastChild());
  }
  return child;
}

ChildRectangle *
PackedRectangle::nextBranch(ChildRectangle *child)
{
  PackedRectangle *parent = (PackedRectangle *) child->getParent();
  if(parent) {
    ChildRectangle *next_branch = parent->nextChild(child);
    if(next_branch)
      return next_branch;
    else
      return nextBranch(parent);
  }
  
  return NULL;
}

ChildRectangle *
PackedRectangle::prevBranch(ChildRectangle *child)
{
  PackedRectangle *parent = (PackedRectangle *) child->getParent();
  if(parent) {
    ChildRectangle *prev_branch = parent->prevChild(child);
    if(prev_branch)
      return prev_branch;
    else
      return prevBranch(parent);
  }

  return NULL;
}

ChildRectangle *
PackedRectangle::nextLeaf(ChildRectangle *leaf) 
{
  if(leaf->numChildren())
    return ((PackedRectangle *) leaf)->firstLeafChild(leaf);

  ChildRectangle *next = ((PackedRectangle *) leaf->getParent())->nextChild(leaf);
  if(next)
    if(next->numChildren())
      return ((PackedRectangle *) next)->nextLeaf(((PackedRectangle *) next)->firstChild());
    else
      return next;

  PackedRectangle *parent = (PackedRectangle *) this->nextBranch(leaf);
  if(parent)
    return parent->firstLeafChild(parent);
  else
    return NULL;
}

ChildRectangle *
PackedRectangle::prevLeaf(ChildRectangle *leaf) 
{
  if(leaf->numChildren())
    return ((PackedRectangle *) leaf)->lastLeafChild(leaf);

  ChildRectangle *prev = ((PackedRectangle *) leaf->getParent())->prevChild(leaf);
  if(prev)
    if(prev->numChildren())
      return ((PackedRectangle *) prev)->prevLeaf(((PackedRectangle *) prev)->lastChild());
    else
      return prev;

  PackedRectangle *parent = (PackedRectangle *) this->prevBranch(leaf);
  if(parent)
    return parent->lastLeafChild(parent);
  else
    return NULL;
}
