#include "VDD.h"
#include "RTECursor.h"
#include "RTESet.h"
#include <stdlib.h>
#include "smemory.h"
#include "RootElement.h"

// static class variables for RTECursor
Coord RTECursor::default_width, RTECursor::default_height, RTECursor::default_offaxis;

RTECursor::RTECursor(PackedRectangle *p, RTESelectionManager *sm)
{
	x = y = drag_x = drag_y = 0;
	parent = p;
	associated_selection_manager = sm;
	this->setSize(w, h);
	this->setPositionNotValid();
}

void 
RTECursor::setSize(Coord &width, Coord &height) 
{
  switch (this->parent->getPrimaryWritingDir()) {
  case LeftToRight: case RightToLeft: // vertical cursor
    width  = default_offaxis;
    height = default_height; // one line high
    break;
  case TopToBottom: case BottomToTop:    // horizontal cursor
    width  = default_width; // one column wide
    height = default_offaxis;
    break;
  case RTENone:             // vertical cursor of default size
    width  = default_offaxis;
    height = default_offaxis; 
    break;
  }
}

int
RTECursor::becomeChildOf(ParentRectangle *p)
{
	const int allow_overlap = TRUE;
	const int redraw = TRUE;
	return this->becomeChildOf(p, redraw, allow_overlap);
}

int
RTECursor::becomeChildOf(ParentRectangle *p, int redraw)
{
	const int allow_overlap = TRUE;
	return this->becomeChildOf(p, redraw, allow_overlap);
}

int
RTECursor::becomeChildOf(ParentRectangle *p, int redraw, int allow_overlap)
{
  // ignore value passed in and always allow overlap for cursor
  
  if(ChildRectangle::becomeChildOf(p, redraw, TRUE)){
    Coord width, height;
    this->setSize(width, height);
    changeSizeTo(width, height);
    return TRUE;
  }	
  
  return FALSE;
}


RTECursor::RTECursor(PackedRectangle *p, RTESelectionManager *sm, 
               Coord px, Coord py, Coord pw, Coord ph, int snap)
{
	drag_x = drag_y = 0;
	w = h = 1;
	parent = p;
	associated_selection_manager = sm;
	if(pw < 0 || ph < 0) halt("negative height or width");
	this->w = pw;  this->h = ph;
	if(px < 0 || py < 0) halt("negative x or y");
}


CursorMode RTECursor::getMode()
{
	return this->mode;
}

void 
RTECursor::setDefaults(Coord width, Coord height, Coord offaxis)
{
	default_width   = width;
	default_height  = height;
	default_offaxis = offaxis;
}

void RTECursor::setInsertMode()
{
	if(this->mode != Insert) {
		this->mode = Insert;
		this->requestRedraw();
	}
}


void RTECursor::setReplaceMode()
{
	if(this->mode != Replace) {
		this->mode = Replace;
		this->requestRedraw();
	}
}


void RTECursor::cycleMode()
{
	// Set mode to next mode, wrapping back to first after last.
	if(this->mode < LAST_CURSOR_MODE)
		this->mode = (CursorMode) ((int)this->mode+1);
	else
		this->mode = FIRST_CURSOR_MODE;
	this->requestRedraw();
}

void RTECursor::positionAt(Coord xp, Coord yp, int snap)
{
	if(xp < 0 || yp < 0) halt("negative xp or yp");
	if(snap) {
		ElementSet *hit = this->parent->getLeafElementsAtLocation(xp, yp);
		Rectangle *r=0;
		switch (hit->numElements()) {
			case 0:
				r = this->parent;
				break;
			case 1:
				r = hit->getElementNumber(0);
				break;
			default:
				// overlapping elements at this location - assume one of
				// them is the cursor (me) and do nothing
				break;
		}
		((PackedRectangle *) this->parent)->alignWithEdge(r, this);
		delete hit; hit = 0;
	}
	drag_x = x;  drag_y = y;
}

void RTECursor::getPosition(Coord *px, Coord *py)
{
	*px = this->x;
	*py = this->y;
}

Coord RTECursor::getSize()
{
	Coord result=0;
	switch (((PackedRectangle *) this->parent)->getPrimaryWritingDir()) {
		case LeftToRight:  
		case RightToLeft:
		case RTENone:
			result = this->w; break;
		case BottomToTop:  
		case TopToBottom:
			result = this->h; break;
	}
	return result;
}


static void projectToBoundary(Coord *x, Coord *y, Coord parent_w, Coord parent_h, Direction dir)
{
	// Project the point (*x, *y) in specified direction to boundary of parent rectangle.
	switch (dir) {
		case LeftToRight: *x = parent_w; break;
		case RightToLeft: *x = 0;        break;
		case TopToBottom: *y = parent_h; break;
		case BottomToTop: *y = 0;        break;
		case RTENone: halt("called with direction RTENone");
	}
}


static int area(Coord x1, Coord y1, Coord x2, Coord y2)
{
	// area of rectangle defined by two diagonally opposite corner points
	return abs((x2-x1) * (y2-y1));
}


static Rectangle *makeRect(Coord x1, Coord y1, Coord x2, Coord y2, Coord size)
{
	// Given two points which have either the same x coord or the same y coord, 
	// make a rectangle with one side defined by the points and the side 
	// perpendicular has length size.

	assert((x1 == x2 || y1 == y2) && size > 0);
	Coord ulx = MIN(x1, x2), uly = MIN(y1, y2), rw, rh;
	if(y1 == y2) {
		rh = size;
		rw = abs(x2 - x1);
	} else {
		rw = size;
		rh = abs(y2 - y1);
	}
	return new Rectangle(ulx, uly, rw, rh);
}

#define SIGN(a) ( (a) >= 0 ? 1 : -1 )
static Coord nearestMultipleAway(Coord x, Coord xx, Coord s)
{
        assert(s > 0 && x >= 0 && xx >= 0);
        int d = abs(xx - x);
        d -= d % s;
        int result = x + d * SIGN(xx - x);
        assert((result >= 0) && (result <= xx) && (abs(result - x) % s == 0));
        return result;
}

static void align(Coord x1, Coord y1, Coord *x2, Coord *y2, Direction scnd_dir, Coord spacing)
{
	// Force (x1,y2)'s distance from (x1,y1) in the secondary spacing direction
	// to be a multiple of the spacing increment.  Round down.

	assert(x1 >= 0 && y1 >= 0 && *x2 >= 0 && *y2 >= 0);
	switch (scnd_dir) {
		case TopToBottom:  case BottomToTop:
			*y2 = nearestMultipleAway(y1, *y2, spacing);
			break;
		case LeftToRight:  case RightToLeft:
			*x2 = nearestMultipleAway(x1, *x2, spacing);
			break;
		case RTENone: halt("called with direction RTENone");
	}
	assert(x1 >= 0 && y1 >= 0 && *x2 >= 0 && *y2 >= 0);
}


static void move(Coord *x, Coord *y, Coord dist, Direction dir)
{
	assert(*x >= 0 && *y >= 0);
	switch (dir) {
		case TopToBottom: *y += dist; break;
		case BottomToTop: *y -= dist; break;
		case LeftToRight: *x += dist; break;
		case RightToLeft: *x -= dist; break;
		case RTENone: halt("called with direction RTENone");
	}
	assert(*x >= 0 && *y >= 0);
}


ChildRectangle *
RTECursor::getGoodChild(ElementSet *set)
{
  int i, size = set->numElements();
  for(i = 0; i < size; i++)
    if(set->getElementNumber(i) != this)
      return set->getElementNumber(i);

  return NULL;
}

void
RTECursor::elementsBoundedBy(Rectangle *area, ElementSet **set, ChildRectangle **first, ChildRectangle **last)
{ 
  RootElement *root = this->parent->getRoot();

  ElementSet *first_set = root->getLeafElementsAtLocation(area->getX(), area->getY()),
             *last_set  = root->getLeafElementsAtLocation(area->getX()+area->getW()-1, area->getY()+area->getH()-1);

  // get the first and last children in the selected set
  *first = getGoodChild(first_set);
  *last  = getGoodChild(last_set);

  ElementSet *es = NULL;
  if(first && last) {
    // determine order
    ChildRectangle *next = *first;
    while(next != *last && !(*last)->isParentOf(next)) {
      next = root->nextLeaf(next);
      if(!next) {
	next   = *first;
	*first = *last;
	*last  = next;
	break;
      }
    }

    if((*first)->numChildren())
      *first = ((PackedRectangle *) (*first))->firstLeafChild(*first);
    if((*last)->numChildren())
      *last  = ((PackedRectangle *) (*last))->lastLeafChild(*last);    

    es = new ElementSet;
    es->addElement(*first);
    if(*first != *last) {
      es->addElement(*last);
      
      ChildRectangle *next = root->nextLeaf(*first);
      while(next && next != *last && !(*last)->isParentOf(next)) {
	es->addElement(next);
	next = root->nextLeaf(next);
      }

      while((*last)->isParentOf(next)) {
	es->addElement(next);
	next = root->nextLeaf(next);
      }

    }
  }

  delete first_set;
  delete last_set;

  *set = es;    
}

void
RTECursor::startDragSelect(Coord sx, Coord sy)
{
  this->drag_x = sx;
  this->drag_y = sy;
  associated_selection_manager->clear();
}

void 
RTECursor::dragSelectTo(Coord sx, Coord sy)
{
	// Toggle selection of the elements in the region from (drag_x, drag_y) to (sx, sy),
	// then update drag_x and drag_y.  This allows incremental selection
	// without re-selecting all the way back to the starting cursor position
	// every time the mouse moves.  It can also handle large select regions.

	if(ABS(sx-drag_x) < 3 && ABS(sy-drag_y) < 3)
	  return;  // null or very small drag region

	sx = (sx < 0) ? 0 : sx;
	sy = (sy < 0) ? 0 : sy;

	Rectangle area(drag_x, drag_y, sx-drag_x+1, sy-drag_y+1);
	ChildRectangle *first, *last;
	ElementSet *es, *cs = new ElementSet;	
        this->elementsBoundedBy(&area, &es, &first, &last);
	this->associated_selection_manager->getSet()->copyTo(cs);
	if(es)
	  cs->removeIntersection(es);
	this->associated_selection_manager->deselect(cs);
	if(es) {
	  this->associated_selection_manager->select(es);
	  this->associated_selection_manager->setFirst(first);
	  this->associated_selection_manager->setLast(last);
	}
	delete es; es = 0;
	delete cs; cs = 0;
}


void RTECursor::endDragSelect()
{
	this->x = this->drag_x;
	this->y = this->drag_y;
	this->requestRedrawBorder();
}


ChildRectangle *RTECursor::getNextElement()
{
	return ((PackedRectangle *) this->parent)->nextChild(this, TRUE);
}


ChildRectangle *RTECursor::getPrevElement()
{
	return ((PackedRectangle *) this->parent)->prevChild(this, TRUE);
}


int RTECursor::draw()
{
	int result = TRUE;

	VirtualDisplayDevice *vdd=0;
	RootElement *r = this->getRoot();
	if(r) vdd = r->getDisplayDevice();

	Coord x_pos, y_pos;
	this->getAbsoluteCoords(&x_pos, &y_pos);

	if(vdd)
	  vdd->fillRect(x_pos, y_pos, w, h);

	return result;  // success
}

