
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"


/* 
// Scott's variant -- It just uses an array 
*/


#include "str-lib.h"
#include <math.h>

#include "rafutil.h"
#include "smemory.h"
#include "svbmap.h"

#define SVBMAP_CHECK TRUE

static int4 svbmapTableSize(svbmap *a)
{
	/* Return number of items in table */
	/* Extra element is allocated at beginning of table
	   and table size stored in its data pointer */

	if (SVBMAP_CHECK && !a) 
		return 0;
	else
		return (int4) a->data; /* segment lost ok */
}

static int4 svbmapSetTableSize(svbmap *a, int4 i)
{
	/* Keep track of number of elements in table */
	/* Extra element is allocated at beginning of table
	   and table size stored in its data pointer */

	if (SVBMAP_CHECK && !a)
		return FALSE;
	else {
		a->data = (void *) i;
		return TRUE;
	}
}

static char *
my_strdup(const char *s)
{
	char *result = smalloc((strlen(s)+1) * sizeof(char));
	if(result)
		strcpy(result, s);
	return result;
}

static void svbmapSetRecord(svbmap *a, const int4 i, const char *key, const void  *data)
{
	/* Put data into i-th array position, where array positions
	   are numbered starting at zero but stored in locations starting
	   at one. Make a copy of the key string. */

	if (SVBMAP_CHECK && !a) 
		halt("Error in svbmapSetRecord: array not allocated");
	a[i+1].name = my_strdup(key);  
	if (SVBMAP_CHECK && !a[i+1].name)
		halt("Error in svbmapSetRecord: strdup failed");
	a[i+1].data = (void*)data;
}

static void svbmapCopyRecord(svbmap *a, int4 i, int4 j)
{
	/* copy data from i-th position to j-th,
	   hiding the fact that the first record
	   is used for array size */

	int4 max = svbmapTableSize(a) - 1;

	if ( SVBMAP_CHECK && ((i < 0) || (i > max) || (j < 0) || (j > max)) )
		halt("Error in svbmapCopyRecord: index out of range");
	else {
		a[j+1].name = a[i+1].name;
		a[j+1].data = a[i+1].data;
	}
}

static int4 svbmapBSearch(svbmap *a, const char *key, int4 *found)
{
	/* return the index of the table record where
	   the specified key should go, where records
	   are numbered starting at 0 , but they occupy
	   locations 1..n of the array a[] */

	const svbmap *items = &a[1];
	int4 start=0, middle, diff, end=svbmapTableSize(a)-1;
	
	diff = strcmp(key, items[start].name);
	if (diff == 0) {
		*found = TRUE;
		return start;
	} else if (diff < 0) 
		/* do nothing */ ;
	else if ((diff > 0) && (end == start)) 
		start = end+1;
        else if (strcmp(key, items[end].name) > 0)
                start = end+1;
        else {
		int4 shift=0;
		start = 0;
		do{
                        middle=(start+end+shift)/2;
                	diff=strcmp(key, items[middle].name);
                	if(diff == 0) {
                        	*found = TRUE;
				return middle;
                	}
                	else if(diff > 0){
                        	if(start == middle) shift++;
                        	start = middle;
                	}
                	else {
                        	if(end == middle) shift++;
                        	end = middle;
                	}
	
                	/* handle special case where middle
                   	jumps between start and end */
                	if(shift == 2) start = end = middle;
	
        	}while(start < end);
	}

	/*  key not found, but should be inserted in position start */
	*found = FALSE;
	return start;
}

static int4 svbmapTableInsert(const int4 i, svbmap **a, const char *key, const void *value)
{
	/* insert new record at position i, where records
	   are numbered starting from 0 but stored in 
	   locations starting at 1 */
	
	int4 j, n=svbmapTableSize(*a);

	if (!(*a = smrealloc(*a, (n+2) * sizeof(svbmap)))) return FALSE;
	if (!svbmapSetTableSize(*a,n+1)) return FALSE;
	for (j=n; j>i; j--)
		svbmapCopyRecord(*a, j-1, j);
	svbmapSetRecord(*a, i, key, value);
	return TRUE;
}


static int4 svbmapInsert(const int4 i, svbmap **that, const char *key, const void *value)
{
	/* Insert a new item into the table.  May need to initialize table
	   if it is empty. */

	if (! *that) {
		/* initialize new table */
		if (i != 0)
			halt("Error in svbmapInsert: inserting in empty table not at position 0");
		if (!(*that = smalloc(2*sizeof(svbmap))))
			return 1;
		if (!svbmapSetTableSize(*that, 1)) return FALSE;
		svbmapSetRecord(*that, 0, key, value);
		return TRUE;
	} else {
		return svbmapTableInsert(i, that, key, value);
	}
} 

static void svbmapDisposeRecord(svbmap *a, int4 i)
{
	/* free the copy of the string key which we allocated */

	if (a[i+1].name){
		smfree(a[i+1].name);
		a[i+1].name = 0;
	} else 
		halt("Error in svbmapDisposeRecord: tried to dispose null string ptr");
}

static int4 svbmapTableDelete(int4 i, svbmap **a)
{
	/* delete the i-th table record, where
	   records are numbered starting from 0 
	   but stored in locations starting at 1 */

	int4 j, n=svbmapTableSize(*a);

	/* free the copy of the string key which we allocated */
	if (SVBMAP_CHECK && ((i < 0) || (i >= n)) )
		halt("Error in svbmapTableDelete: tried to delete record out of range");
	svbmapDisposeRecord(*a, i);

	if (n > 1) {
		for (j=i; j<n-1; j++)
			svbmapCopyRecord(*a, j+1, j);
		if (!(*a = smrealloc(*a, n * sizeof(svbmap)))) return FALSE;
		if (!svbmapSetTableSize(*a,n-1)) return FALSE;
	} else {
		if (SVBMAP_CHECK && (i != 0)) 
			halt("Error in svbmapTableDelete: wrong i for final delete");
		svbmapFree(a);
	}
	return TRUE;
}

void svbmapFree(svbmap **that)
{
	/* Free what remains of the table after last item deleted */

	if (*that) smfree(*that);
	*that = 0;
}

int4 svbmapEmptyP(svbmap *that)
{
	/* See if the table is empty.  Note that svbmapFree must
	   be called after deleting last item in order for this
	   to work properly. */

	return (that == 0);
}

svbmap * svbmapGet(svbmap *that, const char *key)
{
	/* Get table record corresponding to specified key.
	   Use this routine only if you are sure the record
	   is in the table. */

	svbmap *p;

	if (!(p=svbmapSeek(that,key)))
		halt("Error in svbmapGet: tried to get non-existent entry");

	return p;
}

svbmap * svbmapSeek(svbmap *that, const char *key)
{
	/* Look for a record corresponding to the specified key.
	   Return 0 if none found (not necessarily an error) */

	if (!that) /* empty table */
		return 0;
	else {
		int4 found, i = svbmapBSearch(that, key, &found);
		if (found)
			return &that[i+1];
		else
			return 0;
	}
}

int4 svbmapContainsP(svbmap *that, const char *key)
{
	/* See if the table contains a record for the specified key */

	int4 found;
	
	if (!that) return FALSE;
	svbmapBSearch(that, key, &found);
	return found;
}

int4 svbmapSet(svbmap  **that, const char *key, void *value)
{
	/* Create a table record for a new key and set its value.
	   Note that the key must not already be in the table. */

	int4 found=FALSE, i=0;

	if (!value) return 1;
	if (*that) i = svbmapBSearch(*that, key, &found);
	if (!found)
		if (svbmapInsert(i, that, key, value))
			return 0;
		else
			return 1;
	else
		return 1;
}

void svbmapDelete(svbmap  **that, const char *key)
{
	/* Delete record for specified key.  Note that the
	   key must exist in the table. */

	int4 found, i = svbmapBSearch(*that, key, &found);
	if (found) {
		if (!svbmapTableDelete(i, that)) 
			halt("Error in svbmapDelete: error deleting entry");
	}
	else {
		if (SVBMAP_CHECK) halt("Error in svbmapDelete: non-existent key");
	}
}

void svbmapDeleteRecord(svbmap **that, svbmap *delete)
{
	/* Delete a particular record from the table */

	svbmapTableDelete(delete - (*that+1), that);
}


svbmap * svbmapFirst(svbmap *that)
{
	/* Return first record in the table */

	if (!that)
		return 0;
	else
		return &that[1];
}

svbmap * svbmapNext(svbmap *that, svbmap  *prev)
{
	/* Return record which follows a specified record */

	int4 i = prev - &that[1]; 

	if (i >= svbmapTableSize(that)-1)
		return 0;
	else
		return &that[i+2];
}
